#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import copy
import numbers
from collections import defaultdict
from typing import List, Set, Union

import libsbml

from .parameter import ModelParameter, Parameter, ParameterEntry
from .sbmlutil import (_create_global_parameter, _create_local_parameter,
                       getSpeciesByName)
from .species import Species


class Propensity(object):
    def __init__(self):
        self.propensity_dict = {'species': {}, 'parameters': {}}
        self.name = None

    @staticmethod
    def is_valid_propensity(propensity_type) -> bool:
        """checks whether the given propensity_type is valid propensity.

        It recursively checks all subclasses of Propensity until it finds the
        propensity type. Otherwise raise a Type error

        :param propensity_type: Propensity
        :returns bool
        """
        for propensity in Propensity.get_available_propensities():
            if isinstance(propensity_type, propensity):
                return True
        return False

    @staticmethod
    def _all_subclasses(cls):
        """Returns a set of all subclasses of cls (recursively calculated).

        Source:
        https://stackoverflow.com/questions/3862310/how-to-find-all-the-subclasses-of-a-class-given-its-name
        :param cls: A class in the codebase, for example Propensity
        :return: set of all subclasses from cls
        """
        return set(cls.__subclasses__()).union(
            [s for c in cls.__subclasses__() for s in Propensity._all_subclasses(c)])

    @staticmethod
    def get_available_propensities() -> Set:
        return Propensity._all_subclasses(Propensity)

    def _create_sbml_parameter(self, parameter_name, sbml_model, ratelaw, rename_dict = None):
        """Creates an sbml parameter for a parameter of the given name.

        if self.propensity_dict["parameter"]["parameter_name"] is a Parameter,
            creates a global parameter "Parameter.name_Parameter.part_id_Parameter.mechanism"
            where part_id and mechanism can be empty (but _ will always be incldued for uniqueness).
        if self.propensity_dict["parameter"]["parameter_name"] is a Number,
            creates a local parameter "parameter_name".
        rename_dict allows for param.name to be changed to rename_dict[param.name]
        """
        p = self.propensity_dict["parameters"][parameter_name]
        if isinstance(p, ParameterEntry):
            v = p.value
            p_unit = p.unit
            if p_unit == "":
                p_unit = None
            m = p.parameter_key.mechanism 
            if m is None:
                m = ""
            pid = p.parameter_key.part_id
            if pid is None:
                pid = ""

            if rename_dict is None or p.parameter_name not in rename_dict:
                sbml_name = p.parameter_name+"_"+pid+"_"+m
            else:
                sbml_name = rename_dict[p.parameter_name]+"_"+pid+"_"+m

            return _create_global_parameter(sbml_model, sbml_name, v, p_unit)
            
        elif isinstance(p, int) or isinstance(p, float):
            v = p
            if rename_dict is None or parameter_name not in rename_dict:
                sbml_name = parameter_name
            else:
                sbml_name = rename_dict[parameter_name]

            return _create_local_parameter(ratelaw, sbml_name, v)

        else:
            raise TypeError(f"Invalid item in propensity_diction['parameter']: {p}. Only numbers of ParameterEntries accepted.")

    def _check_parameter(self, parameter, allow_None = False, positive = True):
        """A helper function used in setters to set parameters and do type checking."""
        if isinstance(parameter, Parameter) and (parameter.value > 0 or not positive):
            return parameter
        elif isinstance(parameter, numbers.Real) and (parameter > 0 or not positive):
            return parameter
        elif parameter is None and allow_None:
            return parameter
        else:
            if positive:
                raise ValueError(f"Propensity parameters must be Parameters or floats with positive values. Recieved {type(parameter)}.")
            else:
                raise ValueError(f"Propensity parameters must be Parameters or floats. Recieved {type(parameter)}.")

    def _check_species(self, species, allow_None=False):
        """A helper function used in setters to set species and do type checking."""
        if isinstance(species, Species):
            return species
        elif species is None and allow_None:
            return species
        else:
            raise TypeError(f"Propensity expected a Species: received {type(species)}.")

    def pretty_print(self, show_parameters=True, **kwargs):
        txt = self.pretty_print_rate(**kwargs)
        if show_parameters:
            txt += "\n"+self.pretty_print_parameters(**kwargs)
        return txt

    def pretty_print_rate(self, **kwargs):
        raise NotImplementedError("class Propensity is meant to be subclassed!")

    def pretty_print_parameters(self, show_keys = True, **kwargs):
        txt = ""
        for k in self.propensity_dict["parameters"]:
            p = self.propensity_dict["parameters"][k]
            if isinstance(p, Parameter):
                txt += f"  {k}={p.value}"#p.pretty_print(**kwargs)+"\n"
                if isinstance(p, ModelParameter) and show_keys:
                    txt+=f"\n  found_key=(mech={p.found_key.mechanism}, partid={p.found_key.part_id}, name={p.found_key.name}).\n  search_key=(mech={p.search_key.mechanism}, partid={p.search_key.part_id}, name={p.search_key.name})."
                txt+="\n"
            elif p is not None:
                txt += f"  {k}={p}\n"
        return txt

    @property
    def is_reversible(self):
        """By default, Propensities are assumed to NOT be reversible."""
        return False

    @property
    def k_forward(self):
        raise NotImplementedError

    @property
    def k_reverse(self):
        raise NotImplementedError

    def __eq__(self, other):
        if other.__class__ == self.__class__:
            return other.propensity_dict == self.propensity_dict

    @property
    def species(self) -> List:
        """returns the instance variables that are species type."""
        return list(self.propensity_dict['species'].values())

    def create_kinetic_law(self, reaction, reverse_reaction, stochastic, **kwargs):
        raise NotImplementedError("class Propensity is meant to be subclassed!")

    @classmethod
    def from_dict(cls, propensity_dict):
        merged = propensity_dict['parameters']
        merged.update(propensity_dict['species'])
        return cls(**merged)

    def _create_annotation(self, model, propensity_dict_in_sbml, **kwargs):
        """Create simulator specific annotations to write to Kinetic laws or any other
        part of the SBML model object.

        Annotations are used to take advantage of a simulator specific need/feature.
        """
        annotation_string = ''
        # Add your own simulator specific annotations here

        # For the bioscrape simulator:
        # Check if `for_bioscrape` keyword argument has been passed in **kwargs
        if 'for_bioscrape' in kwargs:
            for_bioscrape = kwargs.get('for_bioscrape')
        else:
            for_bioscrape = False

        if for_bioscrape:
            annotation_string = self._create_bioscrape_annotation(propensity_dict_in_sbml)

        return annotation_string

    def _create_bioscrape_annotation(self, propensity_dict_in_sbml):
        """Propensity Annotations are Used to take advantage of Bioscrape Propensity types
        for faster simulation."""
        annotation_dict = defaultdict()
        for param_name, param_value in propensity_dict_in_sbml['parameters'].items():
            annotation_dict[param_name] = param_value

        for species_name, species in propensity_dict_in_sbml['species'].items():
            annotation_dict[species_name] = species

        annotation_dict["type"] = self.name

        annotation_string = "<PropensityType>"
        for k in annotation_dict:
            annotation_string += " "+ str(k) + "=" + str(annotation_dict[k])
        annotation_string += "</PropensityType>"

        # replace strings to match with bioscrape naming convention
        annotation_string = annotation_string.replace('k_forward', 'k', 1)
        # Bioscrape doesn't have the concept of a reversible reaction - so for both the reverse and forward cases
        # we just make the annotation parameter be called 'k'
        annotation_string = annotation_string.replace('k_reverse', 'k', 1)
        return annotation_string

    def _translate_propensity_dict_to_sbml(self, model, ratelaw):
        # get copy of the propensity_dict and fill with sbml names
        propensity_dict_in_sbml = copy.deepcopy(self.propensity_dict)
        for param_name in propensity_dict_in_sbml['parameters'].keys():
            parameter_in_sbml  =self._create_sbml_parameter(param_name, model, ratelaw)
            propensity_dict_in_sbml['parameters'][param_name] = parameter_in_sbml.getId()

        for species_name, species in propensity_dict_in_sbml['species'].items():
            propensity_dict_in_sbml['species'][species_name] = str(species)

        return propensity_dict_in_sbml


class GeneralPropensity(Propensity):
    def __init__(self, propensity_function: str, propensity_species: List[Species], propensity_parameters: List[ParameterEntry]):
        """A class to define a general propensity.

        :param propensity_function: valid propensity formula defined as a string
        :param propensity_species: list of species that are part of the propensity_function
        :param propensity_parameters: list of parameters that are part of the propensity_function
        """
        super(GeneralPropensity, self).__init__()
        self.propensity_function = propensity_function

        if len(propensity_species) > 0 and not all(isinstance(s, Species) for s in propensity_species):
            raise TypeError('propensity_species must be a list of Species!')

        if len(propensity_parameters) > 0 and not all(isinstance(s, ParameterEntry) for s in propensity_parameters):
            raise TypeError('propensity_parameter must be a list of ParameterEntry!')

        for species in propensity_species:
            if str(species) not in self.propensity_function:
                raise ValueError(f'species: {species} must be part of the formula: {self.propensity_function}')

            self.propensity_dict['species'].update({str(species): species})

        for parameter in propensity_parameters:
            if parameter.parameter_name not in self.propensity_function:
                raise ValueError(f'species: {parameter.parameter_name} must be part of the formula: {self.propensity_function}')

            self.propensity_dict['parameters'].update({parameter.parameter_name: parameter.value})

        self.name = 'general'

    def pretty_print_rate(self, **kwargs):
        return self.propensity_function

    def create_kinetic_law(self, model, sbml_reaction, **kwargs):
        """Creates KineticLaw object for SBML using the propensity_function string."""
        ratelaw = sbml_reaction.createKineticLaw()

        propensity_dict_in_sbml = self._translate_propensity_dict_to_sbml(model=model, ratelaw=ratelaw)

        # replacing the species defined in CRN with valid SBML names
        for species_in_crn, species_in_sbml in propensity_dict_in_sbml['species'].items():
            self.propensity_function.replace(species_in_crn, species_in_sbml)

        # replacing the parameters defined in CRN with valid SBML names
        for parameter_in_crn, parameter_in_sbml in propensity_dict_in_sbml['parameters'].items():
            self.propensity_function.replace(parameter_in_crn, parameter_in_sbml)

        math_ast = libsbml.parseL3Formula(self.propensity_function)
        flag = ratelaw.setMath(math_ast)
        if not flag == libsbml.LIBSBML_OPERATION_SUCCESS or math_ast is None:
            raise ValueError("Could not write the rate law for reaction to SBML. Check the propensity functions of reactions.")
        return ratelaw 


class MassAction(Propensity):
    def __init__(self, k_forward: Union[float, ParameterEntry], k_reverse: Union[float, ParameterEntry] = None):
        super(MassAction, self).__init__()
        self.k_forward = k_forward
        self.k_reverse = k_reverse
        self.name = 'massaction'

    @property
    def k_forward(self):
        if isinstance(self._k_forward, Parameter):
            return self._k_forward.value
        else:
            return self._k_forward

    @k_forward.setter
    def k_forward(self, new_k_forward):
        self._k_forward = self._check_parameter(new_k_forward)
        self.propensity_dict['parameters']['k_forward'] = self._k_forward

    @property
    def k_reverse(self):
        if isinstance(self._k_reverse, Parameter):
            return self._k_reverse.value
        else:
            return self._k_reverse

    @k_reverse.setter
    def k_reverse(self, new_k_reverse):
        self._k_reverse = self._check_parameter(new_k_reverse, allow_None=True)
        if self._k_reverse is not None:
            self.propensity_dict['parameters']['k_reverse'] = self._k_reverse

    @property
    def is_reversible(self):
        if self.k_reverse is None:
            return False
        else:
            return True

    def pretty_print_rate(self, **kwargs):
        crn_reaction = kwargs["reaction"]
        reactant_species = {}
        for w_species in crn_reaction.inputs:
            reactant_species[str(w_species.species)] = w_species
        txt = " Kf="+self._get_rate_formula("k_forward", kwargs["stochastic"], reactant_species)
        if self.is_reversible:
            reactant_species = {}
            for w_species in crn_reaction.outputs:
                reactant_species[str(w_species.species)] = w_species
            txt += "\n Kr="+self._get_rate_formula("k_reverse", kwargs["stochastic"], reactant_species)
        return txt

    def create_kinetic_law(self, model, sbml_reaction, stochastic, reverse_reaction=False, **kwargs):

        if 'crn_reaction' in kwargs:
            crn_reaction = kwargs['crn_reaction']
        else:
            raise ValueError('crn_reaction reference is needed for Massaction kinetics!')

        # create a kinetic law for the sbml_reaction
        ratelaw = sbml_reaction.createKineticLaw()


        # translate the internal representation of a propensity to SBML format
        propensity_dict_in_sbml = self._translate_propensity_dict_to_sbml(model=model, ratelaw=ratelaw)

        # set up the forward sbml_reaction
        if not reverse_reaction:
            reactant_species = {}
            for w_species in crn_reaction.inputs:
                species_id = str(w_species.species)
                reactant_species[species_id] = w_species
            param = propensity_dict_in_sbml['parameters']['k_forward']
            propensity_dict_in_sbml['parameters'].pop('k_reverse', None) #remove the other parameter from the propensities
            ratelaw.removeLocalParameter("k_reverse") #if k_reverse is a local parameter, remove it
        # set up a reverse reaction
        elif reverse_reaction:
            reactant_species = {}
            for w_species in crn_reaction.outputs:
                species_id = str(w_species.species)
                reactant_species[species_id] = w_species
            param = propensity_dict_in_sbml['parameters']['k_reverse']
            propensity_dict_in_sbml['parameters'].pop('k_forward', None) #remove the other parameter from the propensities
            ratelaw.removeLocalParameter("k_forward") #if k_forward is a local parameter, remove it

        rate_formula = self._get_rate_formula(param, stochastic, reactant_species)
        # Set the ratelaw to the rateformula
        math_ast = libsbml.parseL3Formula(rate_formula)
        flag = ratelaw.setMath(math_ast)
        if not flag == libsbml.LIBSBML_OPERATION_SUCCESS or math_ast is None:
            raise ValueError("Could not write the rate law for reaction to SBML. Check the propensity functions of reactions.")
        annotation_string = self._create_annotation(model, propensity_dict_in_sbml=propensity_dict_in_sbml, **kwargs)
        sbml_reaction.appendAnnotation(annotation_string)
        return ratelaw

    def _get_rate_formula(self, rate_coeff_name, stochastic, reactant_species) -> str:

        # Create Rate-strings for massaction propensities
        ratestring = rate_coeff_name

        for species_id, weighted_species in reactant_species.items():
            if stochastic:
                ratestring += '*'
                ratestring += f"{species_id}"
                ratestring += '*'
                ratestring += '*'.join(f" ( {species_id} - {i} )" for i in range(1, weighted_species.stoichiometry))
                
                #Remove trailing *
                if ratestring[len(ratestring)-1] == "*":
                    ratestring = ratestring[:-1]
            else:
                if weighted_species.stoichiometry > 1:
                    ratestring += f" * {species_id}^{weighted_species.stoichiometry}"
                else:
                    ratestring += f" * {species_id}"
        return ratestring

class Hill(Propensity):
    def __init__(self, k: float, s1: Species, K: float, n: float, d: Species):
        Propensity.__init__(self)
        self.k = k
        self.s1 = s1
        self.K = K
        self.n = n
        if d is not None:
            self.d = d

    @property
    def k(self):
        if isinstance(self._k, Parameter):
            return self._k.value
        else:
            return self._k

    @k.setter
    def k(self, new_k):
        self._k = self._check_parameter(new_k)
        self.propensity_dict['parameters']['k'] = self._k

    @property
    def K(self):
        if isinstance(self._K, Parameter):
            return self._K.value
        else:
            return self._K

    @K.setter
    def K(self, new_K):
        self._K = self._check_parameter(new_K)
        self.propensity_dict['parameters']['K'] = self._K

    @property
    def n(self):
        if isinstance(self._n, Parameter):
            return self._n.value
        else:
            return self._n
    @n.setter
    def n(self, new_n):
        self._n = self._check_parameter(new_n)
        self.propensity_dict['parameters']['n'] = self._n

    @property
    def s1(self):
        return self._s1

    @s1.setter
    def s1(self, new_s1):
        self._s1 = self._check_species(new_s1)
        self.propensity_dict['species']['s1'] = self.s1

    @property
    def d(self):
        return self._d

    @d.setter
    def d(self, new_d):
        self._d = self._check_species(new_d, allow_None=True)
        self.propensity_dict['species']['d'] = self._d

    def pretty_print_rate(self, show_parameters = True, **kwargs):
        raise NotImplementedError("Propensity class Hill is meant to be subclassed: try HillPositive, HillNegative, ProportionalHillPositive, or ProportionalHillNegative.")

    def create_kinetic_law(self, model, sbml_reaction, stochastic, **kwargs):
        """This code is reused in all Hill Propensity subclasses."""
        if 'reverse_reaction' in kwargs and kwargs['reverse_reaction'] is True:
            raise ValueError('reverse reactions cannot exist for Hill type Propensities!')

        ratelaw = sbml_reaction.createKineticLaw()

        # translate the internal representation of a propensity to SBML format
        propensity_dict_in_sbml = self._translate_propensity_dict_to_sbml(model=model, ratelaw=ratelaw)

        rate_formula = self._get_rate_formula(propensity_dict=propensity_dict_in_sbml)
        # attach simulator specific annotations to the SBML model, if needed
        annotation_string = self._create_annotation(model, propensity_dict_in_sbml, **kwargs)
        sbml_reaction.appendAnnotation(annotation_string)
        # Set the ratelaw to the rateformula
        math_ast = libsbml.parseL3Formula(rate_formula)
        flag = ratelaw.setMath(math_ast)
        if not flag == libsbml.LIBSBML_OPERATION_SUCCESS or math_ast is None:
            raise ValueError("Could not write the rate law for reaction to SBML. Check the propensity functions of reactions.")

        return ratelaw

    def _get_rate_formula(self, propensity_dict):
        raise NotImplementedError('Hill does not have a rate formula! Check out the subclasses.')


class HillPositive(Hill):
    def __init__(self, k: float, s1: Species, K: float, n: float):
        """ Hill positive propensity is a nonlinear propensity with the following formula.

            p(s1; k, K, n) = k*s1^n/(s1^n + K)

        :param k: rate constant (float)
        :param s1: species (chemical_reaction_network.species)
        :param K: dissociation constant (float)
        """
        Hill.__init__(self=self, k=k, s1=s1, K=K, n=n, d=None)
        self.name = 'hillpositive'

    def pretty_print_rate(self, show_parameters = True, **kwargs):
        return f' Kf = k {self.s1.pretty_print(**kwargs)}^n / ( K^n + {self.s1.pretty_print(**kwargs)}^n )'



    def _get_rate_formula(self, propensity_dict):
        k = propensity_dict['parameters']['k']
        n = propensity_dict['parameters']['n']
        K = propensity_dict['parameters']['K']
        s1 = propensity_dict['species']['s1']
        rate_formula = f"{k}*{s1}^{n} / ( {K}^{n} + {s1}^{n} )"
        return rate_formula


class HillNegative(Hill):
    def __init__(self, k: float, s1: Species, K: float, n: float):
        """ Hill negative propensity is a nonlinear propensity with the following formula.

            p(s1; k, K, n) = k*1/(s1^n + K)

        :param k: rate constant (float)
        :param s1: species (chemical_reaction_network.species)
        :param K: dissociation constant (float)
        :param n: cooperativity (float)
        """
        Hill.__init__(self = self, k=k, s1=s1, K=K, n=n, d=None)
        self.name = 'hillnegative'

    def pretty_print_rate(self, show_parameters = True, **kwargs):
        return f' Kf = k / ( 1 + ({self.s1.pretty_print(**kwargs)}/K)^n )'

    def _get_rate_formula(self, propensity_dict):
        k = propensity_dict['parameters']['k']
        n = propensity_dict['parameters']['n']
        K = propensity_dict['parameters']['K']
        s1 = propensity_dict['species']['s1']
        rate_formula = f"{k} / ( 1 + ({s1}/{K})^{n} )"
        return rate_formula


class ProportionalHillPositive(HillPositive):
    def __init__(self, k: float, s1:Species, K: float, n: float, d:Species):
        """ proportional Hill positive propensity with the following formula.

            p(s1, d; k, K, n) = k*d*s1^n/(s1^n + K)

        :param k: rate constant (float)
        :param s1: species (chemical_reaction_network.species)
        :param K: dissociation constant (float)
        :param n: cooperativity (float)
        :param d: species (chemical_reaction_network.species)
        """
        Hill.__init__(self=self, k=k, s1=s1, K=K, n=n, d=d)
        self.name = 'proportionalhillpositive'

    def pretty_print_rate(self, show_parameters = True,  **kwargs):
        return f' Kf = k {self.d.pretty_print(**kwargs)} {self.s1.pretty_print(**kwargs)}^n / ( K^n + {self.s1.pretty_print(**kwargs)}^n )'


    def _get_rate_formula(self, propensity_dict):
        k = propensity_dict['parameters']['k']
        n = propensity_dict['parameters']['n']
        K = propensity_dict['parameters']['K']
        s1 = propensity_dict['species']['s1']
        d = propensity_dict['species']['d']
        return f"{k}*{d}*{s1}^{n} / ( {K}^{n} + {s1}^{n} )"


class ProportionalHillNegative(HillNegative):
    def __init__(self, k: float, s1: Species, K: float, n: float, d: Species):
        """ proportional Hill negative propensity with the following formula.

            p(s1, d; k, K, n) = k*d/(s1^n + K)

        :param k: rate constant (float)
        :param s1: species (chemical_reaction_network.species)
        :param K: dissociation constant (float)
        :param n: cooperativity (float)
        :param d: species (chemical_reaction_network.species)
        """
        Hill.__init__(self=self, k=k, s1=s1, K=K, n=n, d=d)
        self.name = 'proportionalhillnegative'

    def pretty_print_rate(self, show_parameters=True, **kwargs):
        return f' Kf = k {self.d.pretty_print(**kwargs)} / ( 1 + ({self.s1.pretty_print(**kwargs)}/K)^{self.n} )'

    def _get_rate_formula(self, propensity_dict):
        k = propensity_dict['parameters']['k']
        n = propensity_dict['parameters']['n']
        K = propensity_dict['parameters']['K']
        s1 = propensity_dict['species']['s1']
        d = propensity_dict['species']['d']
        return f"{k}*{d} / ( 1 + ({s1}/{K})^{n} )"




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import copy
from typing import List, Union
from warnings import resetwarnings, warn

from .chemical_reaction_network import ChemicalReactionNetwork
from .component import Component
from .global_mechanism import GlobalMechanism
from .mechanism import Mechanism
from .parameter import ParameterDatabase
from .reaction import Reaction
from .species import Species
from .utils import remove_bindloc


class Mixture(object):
    def __init__(self, name="", mechanisms=None, components=None, parameters=None, parameter_file=None,
                 global_mechanisms=None, species=None, initial_condition_dictionary=None, \
                 global_component_enumerators=None,global_recursion_depth=4, local_recursion_depth = None, **kwargs):
        """A Mixture object holds together all the components (DNA,Protein, etc), mechanisms (Transcription, Translation),
        and parameters related to the mixture itself (e.g. Transcription rate). Default components and mechanisms can be
        added as well as global mechanisms that impacts all species (e.g. cell growth).

        :param name: Name of the mixture
        :param mechanisms: Dictionary of mechanisms
        :param components: List of components in the mixture (list of Components)
        :param parameters: Dictionary of parameters (check parameters documentation for the keys)
        :param parameter_file: Parameters can be loaded from a parameter file
        :param default_mechanisms:
        :param global_mechanisms: dict of global mechanisms that impacts all species (e.g. cell growth)
        """
        # Initialize instance variables
        self.name = name  # Save the name of the mixture

        #recursion depth for global component enumeration
        self.global_recursion_depth = global_recursion_depth

        if(local_recursion_depth is None):
            self.local_recursion_depth = self.global_recursion_depth+2

        #

        # process the components
        if components is None and not hasattr(self, "_components"):
            self.components = []
        else:
            self.add_components(components)

        # process mechanisms:
        if mechanisms is None and not hasattr(self, "_mechanisms"):
            self.mechanisms = {}
        else:
            self.add_mechanisms(mechanisms)

        # process global_mechanisms:
        # Global mechanisms are applied just once ALL species generated from
        # components inside a mixture
        # Global mechanisms should be used rarely, and with care. An example
        # usecase is degradation via dilution.
        if global_mechanisms is None and not hasattr(self, "_global_mechanisms"):
            self.global_mechanisms = {}
        elif global_mechanisms is not None:
            self.add_mechanisms(global_mechanisms)

        # global component enumerators
        if global_component_enumerators is None:
            self.global_component_enumerators = []
        else:
            self.global_component_enumerators = global_component_enumerators

        # process the species
        self.add_species(species)

        # Create a paraemter database
        self.parameter_database = ParameterDatabase(parameter_file = parameter_file, parameter_dictionary = parameters, **kwargs)
        
        # CRN is stored here during compilation
        self.crn = None

    def add_species(self, species: Union[List[Species], Species]):
        if not hasattr(self, "added_species"):
            self.added_species = []

        if species is not None:
            if not isinstance(species, list):
                species_list = [species]
            else:
                species_list = species

            assert all(isinstance(x, Species) for x in species_list), 'only Species type is accepted!'

            self.added_species += species_list

    def set_species(self, species: Union[Species, str], material_type=None, attributes=None):
        """Used to set internal species from strings, Species or Components

        :param species: name of a species or a species instance
        :param material_type: material type of a species as a string
        :param attributes: Species attribute
        :return: Species in the mixture
        """
        if isinstance(species, Species):
            return species
        elif isinstance(species, str):
            return Species(name=species, material_type=material_type, attributes=attributes)
        elif isinstance(species, Component) and species.get_species() is not None:
            return species.get_species()
        else:
            raise ValueError("Invalid Species: string, chemical_reaction_network.Species or Component with implemented .get_species() required as input.")

    @property
    def components(self):
        return self._components

    @components.setter
    def components(self, components):
        self._components = []
        self.add_components(components)
    
    def add_component(self, component):
        """this function adds a single component to the mixture."""
        if not hasattr(self, "_components"):
            self.components = []

        if isinstance(component, list):
            self.add_components(component)
        else:
            assert isinstance(component, Component), "the object: %s passed into mixture as component must be of the class Component" % str(component)

            # Check if component is already in self._components
            for comp in self._components:
                if type(comp) == type(component) and comp.name == component.name:
                    raise ValueError(f"{comp} of the same type and name already in Mixture!")
            else:
                # Components are copied before being added to Mixtures
                component_copy = copy.deepcopy(component)
                component_copy.set_mixture(self)
                self.components.append(component_copy)

    def get_mechanism(self, mechanism_type):
        """Searches the Mixture for a Mechanism of the correct type.

        If no Mechanism is found, None is returned.
        """
        if not isinstance(mechanism_type, str):
            raise TypeError(f"mechanism_type must be a string. Recievied {mechanism_type}.")

        if mechanism_type in self.mechanisms:
            return self.mechanisms[mechanism_type]
        else:
            return None

    def add_components(self, components: Union[List[Component], Component]):
        """This function adds a list of components to the mixture.
        """
        if isinstance(components, Component):
            self.add_component(components)
        elif isinstance(components, List):
            for component in components:
                self.add_component(component)
        else:
            raise ValueError(f"add_components expected a list of Components. Received {components}")

    def get_component(self, component=None, name=None, index=None):
        """Function to get components from Mixture._components.

        One of the 3 keywords must not be None.

        :param component: an instance of a component. Searches Mixture._components for a Component with the same type and name.
        :param name: str. Searches Mixture._components for a Component with the same name
        :param index: int. returns Mixture._components[index]
        :return: if nothing is found, returns None.
        """
        if [component, name, index].count(None) != 2:
            raise ValueError(f"get_component requires a single keyword. Received component={component}, name={name}, index={index}.")
        if not (isinstance(component, Component) or component is None):
            raise ValueError(f"component must be of type Component. Received {component}.")
        if not (isinstance(name, str) or name is None):
            raise ValueError(f"name must be of type str. Received {name}.")
        if not (isinstance(index, int) or index is None):
            raise ValueError(f"index must be of type int. Received {index}.")

        matches = []
        if index is not None:
            matches.append(self.components[index])
        else:
            for comp in self.components:
                if component is not None:
                    if type(comp) == type(component) and comp.name == component.name:
                        matches.append(comp)
                elif name is not None:
                    if comp.name == name:
                        matches.append(comp)
        if len(matches) == 0:
            return None
        elif len(matches) == 1:
            return matches[0]
        else:
            warn("get_component found multiple matching components. A list has been returned.")
            return matches 

    @property
    def mechanisms(self):
        """mechanisms stores Mixture Mechanisms."""
        return self._mechanisms

    @mechanisms.setter
    def mechanisms(self, mechanisms):
        self._mechanisms = {}
        self.add_mechanisms(mechanisms, overwrite=True)
        
    def add_mechanism(self, mechanism, mech_type=None, overwrite=False):
        """adds a mechanism of type mech_type to the Mixture mechanism_dictionary.

        :param mechanism: a Mechanism instance
        :param mech_type: the type of mechanism. defaults to mechanism.mech_type if None
        :param overwrite: whether to overwrite existing mechanisms of the same type (default False)
        :return:
        """
        if not hasattr(self, "_mechanisms"):
            self._mechanisms = {}

        if not isinstance(mechanism, Mechanism):
            raise TypeError(f"mechanism must be a Mechanism. Received {mechanism}.")

        if mech_type is None:
            mech_type = mechanism.mechanism_type
        if not isinstance(mech_type, str):
            raise TypeError(f"mechanism keys must be strings. Received {mech_type}")

        if isinstance(mechanism, GlobalMechanism):
            self.add_global_mechanism(mechanism, mech_type, overwrite)
        elif isinstance(mechanism, Mechanism):
            if mech_type in self._mechanisms and not overwrite:
                raise ValueError(f"mech_type {mech_type} already in Mixture {self}. To overwrite, use keyword overwrite = True.")
            else:
                self._mechanisms[mech_type] = copy.deepcopy(mechanism)
        
    def add_mechanisms(self, mechanisms, overwrite=False):
        """This function adds a list or dictionary of mechanisms to the mixture.

        Can take both GlobalMechanisms and Mechanisms

        :param mechanisms: a Mechanism instance
        :param overwrite: whether to overwrite existing mechanisms of the same type (default False)
        :return:
        """
        if isinstance(mechanisms, Mechanism):
            self.add_mechanism(mechanisms, overwrite = overwrite)
        elif isinstance(mechanisms, dict):
            for mech_type in mechanisms:
                self.add_mechanism(mechanisms[mech_type], mech_type, overwrite = overwrite)
        elif isinstance(mechanisms, list):
            for mech in mechanisms:
                self.add_mechanism(mech, overwrite = overwrite)
        else:
            raise ValueError(f"add_mechanisms expected a list of Mechanisms. Recieved {mechanisms}")

                
    @property
    def global_mechanisms(self):
        """global_mechanisms stores global Mechanisms in the Mixture."""
        return self._global_mechanisms

    @global_mechanisms.setter
    def global_mechanisms(self, mechanisms):
        self._global_mechanisms = {}
        if isinstance(mechanisms, dict):
            for mech_type in mechanisms:
                self.add_global_mechanism(mechanisms[mech_type], mech_type, overwrite = True)
        elif isinstance(mechanisms, list):
            for mech in mechanisms:
                self.add_global_mechanism(mech, overwrite = True)

    def add_global_mechanism(self, mechanism, mech_type = None, overwrite = False):
        """adds a mechanism of type mech_type to the Mixture global_mechanism dictonary.

        keywordS:
          mechanism: a Mechanism instance
          mech_type: the type of mechanism. defaults to mechanism.mech_type if None
          overwrite: whether to overwrite existing mechanisms of the same type (default False)
        """
        if not hasattr(self, "_global_mechanisms"):
            self._global_mechanisms = {}

        if not isinstance(mechanism, GlobalMechanism):
            raise TypeError(f"mechanism must be a GlobalMechanism. Recieved {mechanism}.")

        if mech_type is None:
            mech_type = mechanism.mechanism_type
        if not isinstance(mech_type, str):
            raise TypeError(f"mechanism keys must be strings. Recieved {mech_type}")

        if mech_type in self._mechanisms and not overwrite:
            raise ValueError(f"mech_type {mech_type} already in Mixture {self}. To overwrite, use keyword overwrite = True.")
        else:
            self._global_mechanisms[mech_type] = copy.deepcopy(mechanism)

    def update_parameters(self, parameter_file = None, parameters = None, overwrite_parameters = True):
        if parameter_file is not None:
            self.parameter_database.load_parameters_from_file(parameter_file, overwrite_parameters = overwrite_parameters)

        if parameters is not None:
            self.parameter_database.load_parameters_from_dictionary(parameters, overwrite_parameters = overwrite_parameters)
    
    def get_parameter(self, mechanism, part_id, param_name):
        param = self.parameter_database.find_parameter(mechanism, part_id, param_name)
        return param

    def get_initial_concentration(self, S: Union[List, Species], component=None):
        """
        Tries to find an initial condition of species s using the parameter hierarchy using the key:

        1. Searches Component's ParameterDatabase using the key:
            mechanisms = "initial concentration"
            part_id = mixture.name
            parameter_name = str(s)

            if s == component.get_species, also checks with parameter_name=component.name

        2. Searches the Mixture's ParameterDatabase using the key:
            mechanisms = "initial concentration"
            part_id = mixture.name
            parameter_name = str(s)

            if s == component.get_species, also checks with parameter_name=component.name

        3. Defaults to 0
        """
        if isinstance(S, Species):
            S = [S]

        #flatten the species list
        S = Species.flatten_list(S)

        init_conc_dict = {}
        for s in S:
            if not isinstance(s, Species):
                raise ValueError(f"{s} is not a Species! Can only find initial concentration of a Species.")
            init_conc = None
            #1 Check the component
            if component is not None:
                init_conc = component.get_parameter(param_name = str(s), part_id = self.name, mechanism = "initial concentration", check_mixture = False, return_none = True)

                if init_conc is None and component.get_species() == s:
                    init_conc = component.get_parameter(param_name = component.name, part_id = self.name, mechanism = "initial concentration", check_mixture = False, return_none = True)
                
            #2 Check self
            if init_conc is None:
                init_conc = self.get_parameter(param_name = str(s), part_id = self.name, mechanism = "initial concentration")

                if init_conc is None and component is not None and component.get_species() == s:
                    init_conc = self.get_parameter(param_name = component.name, part_id = self.name, mechanism = "initial concentration")

            if init_conc is None:
                init_conc = 0

            init_conc_dict[s] = init_conc

        return init_conc_dict

    def add_species_to_crn(self, new_species, component = None, no_initial_concentrations = False, copy_species = True):

        if self.crn is None:
            self.crn = ChemicalReactionNetwork(species = [], reactions = [])

        if isinstance(new_species, Species):
            new_species = [new_species]

        self.crn.add_species(new_species, copy_species = copy_species)

        if not no_initial_concentrations:
            init_conc_dict = self.get_initial_concentration(remove_bindloc(new_species), component)
            self.crn.initial_concentration_dict = init_conc_dict

    def apply_global_mechanisms(self, species) -> (List[Species], List[Reaction]):
        # update with global mechanisms

        global_mech_species = []
        global_mech_reactions = []
        if self.global_mechanisms:
            for mech in self.global_mechanisms:
                # Update Global Mechanisms
                global_mech_species += self.global_mechanisms[mech].update_species_global(species, self)
                global_mech_reactions += self.global_mechanisms[mech].update_reactions_global(species, self)

        self.add_species_to_crn(global_mech_species, component = None)
        self.crn.add_reactions(global_mech_reactions)

    def component_enumeration(self, comps_to_enumerate = None,recursion_depth=10) -> List[Component]:
        #Components that produce components through Component Enumeration

        all_components = []
        new_components = []
        if(comps_to_enumerate is None):
            comps_to_enumerate = list(self.components)

        if recursion_depth == 0:
            all_components =  comps_to_enumerate
        else:
            #Recursion depth
            for a in range(recursion_depth):
                for component in comps_to_enumerate:
                    component.set_mixture(self)
                    enumerated = component.enumerate_components(previously_enumerated=all_components+new_components)
                    new_components += enumerated

                all_components += comps_to_enumerate
                comps_to_enumerate = list(new_components)
                new_components = []

        if(len(comps_to_enumerate) > 0):
            warn("Mixture was left with unenumerated components "+str(', '.join([str(c) for c in comps_to_enumerate])))
        return all_components
    
    def global_component_enumeration(self,comps_to_enumerate=None, recursion_depth=None) -> List[Component]:
        """components that produce other components infinitely"""
        if(recursion_depth is None):
            recursion_depth = self.global_recursion_depth

        if(comps_to_enumerate is None):
            comps_to_enumerate = list(self.components) #these go into the ComponentEnuemrators

        #Recursion depth
        enumerated_components = list(comps_to_enumerate) #These will be returned
        for global_enumerator in self.global_component_enumerators:
            global_enumerator.reset(enumerated_components) #reset the enumeration if there's any stored info
        for a in range(recursion_depth):
            new_comps_to_enumerate = [] #these will be added to comps_to_enumerate at the end of the iteration

            for global_enumerator in self.global_component_enumerators:
                enumerated = global_enumerator.enumerate_components(comps_to_enumerate,\
                                            previously_enumerated=enumerated_components) #this should be only NEWLY CREATED components
                for c in enumerated:
                    #These components are passed into the enumerator next recursion
                    if c not in new_comps_to_enumerate and c not in comps_to_enumerate:
                        new_comps_to_enumerate.append(c)

                    #These components are returend
                    if c not in enumerated_components:
                        enumerated_components.append(c)

            #Update comps_to_enumerate
            comps_to_enumerate += new_comps_to_enumerate

        return enumerated_components

    def compile_crn(self, recursion_depth = None, initial_concentration_dict = None, return_enumerated_components = False,
        initial_concentrations_at_end = False, copy_objects = True, add_reaction_species = True) -> ChemicalReactionNetwork:
        """Creates a chemical reaction network from the species and reactions associated with a mixture object.
        :param initial_concentration_dict: a dictionary to overwride initial concentrations at the end of compile time
        :param recursion_depth: how deep to run the Local and Global Component Enumeration
        :param return_enumerated_components: returns a list of all enumerated components along with the CRN
        :param initial_concentrations_at_end: if True does not look in Components for Species' initial concentrations and only checks the Mixture database at the end.
        :param copy_objects: Species and Reactions will be copied when placed into the CRN. Protects CRN validity at the expense of compilation speed.
        :param add_reaction_species: Species inside reactions will be added to the CRN. Ensures no missing species as teh expense of compilation speed.
        :return: ChemicalReactionNetwork
        """
        resetwarnings()#Reset warnings - better to toggle them off manually.

        #Create a CRN to filter out duplicate species
        self.crn = ChemicalReactionNetwork([], [])

        #add the extra species to the CRN
        self.add_species_to_crn(self.added_species, component = None, no_initial_concentrations = initial_concentrations_at_end, copy_species = copy_objects)
        
        #get the recursion depth
        if(recursion_depth is None):
            recursion_depth = self.global_recursion_depth
        #Run global enumeration
        globally_enumerated_components = self.global_component_enumeration(recursion_depth=recursion_depth)
        #Run Local Enumeraton
        enumerated_components = self.component_enumeration(globally_enumerated_components, recursion_depth = self.local_recursion_depth) #This includes self.components 
        #reset the Components' mixture to self - in case they have been added to other Mixtures
        for c in enumerated_components:
            c.set_mixture(self)

        #Append Species from each Component
        species = []
        for component in enumerated_components:
            self.add_species_to_crn(component.update_species(), component, no_initial_concentrations = initial_concentrations_at_end, copy_species = copy_objects)

        #Append Reactions from each Component
        for component in enumerated_components:
            self.crn.add_reactions(component.update_reactions(), copy_reactions = copy_objects, add_species =  add_reaction_species)


        #global mechanisms are applied last and only to all the species
        #the reactions and species are added to the CRN
        self.apply_global_mechanisms(self.crn._species)

        if initial_concentrations_at_end:
            self.crn.initial_concentration_dict = self.get_initial_concentration(self.crn._species, component = None)
        #Manually change/override initial conditions at compile time
        if initial_concentration_dict is not None:
            self.crn.initial_concentration_dict = initial_concentration_dict

        if return_enumerated_components:
            return self.crn, enumerated_components
        else:
            return self.crn

    def __str__(self):
        return type(self).__name__ + ': ' + self.name

    def __repr__(self):
        txt = str(self)+"\n"
        if self.components:
            txt += "Components = ["
            for comp in self.components:
                txt+="\n\t"+str(comp)
        if self.mechanisms:
            txt+=" ]\nMechanisms = {"
            for mech in self.mechanisms:
                txt+="\n\t"+mech+":"+self.mechanisms[mech].name
        if self.global_mechanisms:
            txt+=" }\nGlobal Mechanisms = {"
            for mech in self.global_mechanisms:
                txt+="\n\t"+mech+":"+self.global_mechanisms[mech].name
        txt+=" }"
        return txt



################################################################
#       DNA_construct: a higher level for construct compilation
#       Author: Andrey Shur
#       Latest update: 7/31/2020
#       Copyright (c) 2020, Build-A-Cell. All rights reserved.
#       See LICENSE file in the project root directory for details.
#
################################################################
import itertools as it
import numbers

from typing import Dict, Union
from .species import WeightedSpecies, Species
from .parameter import Parameter
from warnings import warn

def all_comb(input_list):
    out_list = []
    for i in range(1,len(input_list)+1):
        out_list += it.combinations(input_list,i)
    return out_list


def rev_dir(dir):
    reversedict = {"forward":"reverse","reverse":"forward"}
    return reversedict[dir]

def recursive_parent(s):
    #Recursively goes through Species and gets the top level parent
    if hasattr(s, "parent") and s.parent is not None:
        return recursive_parent(s.parent)
    else:
        return s

def remove_bindloc(spec_list):
        """go through every species on a list and remove any "bindloc" attributes. This is used
        to convert monomers with a parent polymer into the correct species after combinatorial binding
        in things like DNAassembly and RNAassembly."""
        if not isinstance(spec_list, list):
            spec_list = [spec_list]

        out_sp_list = []
        for s in spec_list:
            #go through the species and replace species with their parents, recursively

            if isinstance(s, WeightedSpecies):
                parent = recursive_parent(s.species)
                s.species = parent
                out_sp_list.append(s)
            else:
                parent = recursive_parent(s)
                out_sp_list.append(parent)
                
        return out_sp_list

#Converts a parameter to its Value
def parameter_to_value(p):
    if isinstance(p, Parameter):
        return p.value
    else:
        return p

#Converts a dictionary of Species (or strings) --> Parameters (or Numbers) to Strings --> Numbers
def process_initial_concentration_dict(initial_concentration_dict: Dict[Union[str, Species], Union[numbers.Real, Parameter]]) -> Dict[str, numbers.Real]:
    return {str(key): parameter_to_value(value) for (key, value) in initial_concentration_dict.items()}

#combines two dictionaries by adding missing keys and appending values to existing keys so the
#output dictionary has all the data of both inputs. ONLY WORKS FOR DICTIONARIES WHERE THE VALUES ARE LISTS!!
def combine_dictionaries(dict1,dict2):
        """append lists that share the same key, and add new keys
        WARNING: this only works if the dictionaries have values that are lists"""
        outdict = dict1
        for key in dict2:
            if key in outdict:
                assert(isinstance(dict2[key],list))
                assert(isinstance(outdict[key],list))
                outdict[key] += dict2[key]
            else:
                outdict[key] = dict2[key]
        return outdict
    
def member_dictionary_search(member,dictionary):
    """searches through a dictionary for keys relevant to the given data member.
    Order of returning:
    repr
    name
    material_type
    propensity name
    propensity partid
    propensity mechanism
    """
    if dictionary is None:
        return None
    elif(repr(member) in dictionary):
        return dictionary[repr(member)]
    elif(hasattr(member,"name") and member.name in dictionary):
        return dictionary[member.name]
    elif(hasattr(member,"integrase") and hasattr(member,"site_type") and (str(member.integrase.name),member.site_type) in dictionary):
        return dictionary[(str(member.integrase.name),member.site_type)]
    elif(hasattr(member,"site_type") and member.site_type in dictionary):
        return dictionary[member.site_type]
    elif(hasattr(member,"assembly")):
        typename = type(member).__name__
        if(typename in dictionary):
            return dictionary[typename]
    elif(hasattr(member,"material_type") and hasattr(member,"attributes") and \
                    (member.material_type, tuple(member.attributes)) in dictionary):
        return dictionary[(member.material_type, tuple(member.attributes))]
    elif(hasattr(member,"material_type") and member.material_type in dictionary):
        return dictionary[member.material_type]
    elif(hasattr(member,"attributes") and tuple(member.attributes) in dictionary):
        return dictionary[tuple(member.attributes)]
    elif(hasattr(member,"propensity_type")):
        out_value = None
        try:
            for k,p in member.propensity_type.propensity_dict["parameters"].items():
                if(hasattr(p,"search_key")):
                    mech_str = repr(p.search_key.mechanism).strip('\'\"')
                    partid_str = repr(p.search_key.part_id).strip('\'\"')
                    name_str = repr(p.search_key.name).strip('\'\"')
                    cur_value = out_value
                    if(name_str in dictionary):
                        cur_value = dictionary[name_str] #name of the mechanism that made the reaction
                    if(partid_str in dictionary):
                        cur_value = dictionary[partid_str] #partid used to make the reaction
                    if(mech_str in dictionary):
                        cur_value = dictionary[mech_str] #the type of mechanism that made the reaction
                    if(out_value is not None and cur_value != out_value):
                        warn(f"dictionary search output was {out_value} but now it will be {cur_value}")
                    out_value = cur_value

        except KeyError:
            pass
        return out_value
    return None




#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import copy

from .dna_part import DNA_part
from .species import Species
from .components_basic import DNA, RNA

class RBS(DNA_part):
    """
    A simple RBS class with no regulation. Must be included in a DNAconstruct or DNAassembly to do anything.
    """
    def __init__(self, name: str, assembly=None,
                 transcript=None, protein=None, length=0,
                 mechanisms=None, parameters=None, **keywords):
        self.assembly = assembly
        self.length = length

        DNA_part.__init__(self, name = name, mechanisms = mechanisms,
                           parameters = parameters, **keywords)

        if transcript is None and assembly is None:
            self.transcript = None
        elif transcript is None:
            self.transcript = Species(assembly.name, material_type = "rna")
        else:
            self.transcript = self.set_species(transcript, material_type = "rna")
        
        if protein is None and assembly is not None:
            self.protein = Species(assembly.name, material_type = "protein")
        elif protein is None and assembly is None:
            self.protein = None
        else:
            self.protein = self.set_species(protein, material_type = "protein")

    def update_species(self):
        mech_tl = self.get_mechanism('translation')
        species = []
        species += mech_tl.update_species(transcript = self.transcript, protein = self.protein, component = self, part_id = self.name)
        return species

    def update_reactions(self):
        mech_tl = self.get_mechanism('translation')
        reactions = []

        if self.protein is not None:
            reactions += mech_tl.update_reactions(transcript = self.transcript, protein = self.protein, component = self, part_id = self.name)
        return reactions
    def update_component(self,internal_species=None,**keywords):
        """returns a copy of this component, except with the proper fields updated"""
        if(isinstance(self.parent,DNA)):
            return None
        elif(isinstance(self.parent,RNA)):
            if(self.direction=="forward"):
                out_component = copy.copy(self)
                out_component.transcript = internal_species
                return out_component
            elif(self.direction=="reverse"):
                return None
            else:
                raise AttributeError(f"Unknown direction {self.direction} encountered in {self}")
        else:
            return None
        return out_component

    @classmethod
    def from_rbs(cls, name, assembly, transcript, protein):
        """Helper function to initialize a rbs instance from another rbs or str.

        :param name: either string or an other rbs instance
        :param assembly:
        :param transcript:
        :param protein:
        :return: RBS instance
        """
        if isinstance(name, RBS):
            rbs_instance = copy.deepcopy(name)
            rbs_instance.assembly = assembly
            rbs_instance.transcript = transcript
            rbs_instance.protein = protein
        elif isinstance(name, str):
            rbs_instance = cls(name=name, assembly=assembly,
                               transcript=transcript, protein=protein)
        else:
            raise TypeError(f'RBS can be initialized from string or another RBS! We got {type(name)}')
        return  rbs_instance



# sbmlutil.py - libsbml helper functions
# RMM, 14 Aug 2018
#
# Copyright (c) 2018, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import logging
import warnings
from random import randint
from typing import List
from warnings import warn
from .utils import parameter_to_value
from .units import create_new_unit_definition
import libsbml

# Reaction ID number (global)
reaction_id = 0

logger = logging.getLogger(__name__)


def create_sbml_model(compartment_id="default", time_units='second', extent_units='mole', substance_units='mole',
                      length_units='metre', area_units='square_metre', volume_units='litre', volume=1e-6, model_id=None, **kwargs):
    """Creates an SBML Level 3 Version 2 model with some fixed standard settings.
    Refer to python-libsbml for more information on SBML API.
    :param compartment_id:
    :param time_units:
    :param extent_units:
    :param substance_units:
    :param length_units:
    :param area_units:
    :param volume_units:
    :param volume:
    :param model_id:
    :return:  the SBMLDocument and the Model object as a tuple
    """
    document = libsbml.SBMLDocument(3, 2)
    model = document.createModel()
    if model_id is None:
        model_id = 'biocrnpyler_'+str(randint(1, int(1e6)))
    model.setId(model_id)
    model.setName(model_id)
    # Define units for area (not used, but keeps COPASI from complaining)
    unitdef = model.createUnitDefinition()
    unitdef.setId('square_metre')
    unit = unitdef.createUnit()
    unit.setKind(libsbml.UNIT_KIND_METRE)
    unit.setExponent(2)
    unit.setScale(0)
    unit.setMultiplier(1)

    # Set up required units and containers
    model.setTimeUnits(time_units)  # set model-wide time units
    model.setExtentUnits(extent_units)  # set model units of extent
    model.setSubstanceUnits(substance_units)  # set model substance units
    model.setLengthUnits(length_units)  # area units (never used?)
    model.setAreaUnits(area_units)  # area units (never used?)
    model.setVolumeUnits(volume_units)  # default volume unit

    return document, model


# Creates an SBML id from a chemical_reaction_network.species object
def valid_sbml_id(given_id, document=None):
    # Construct the species ID
    all_ids = []
    if document:
        all_ids = getAllIds(document.getListOfAllElements())
    trans = SetIdFromNames(all_ids)
    valid_id = trans.getValidIdForName(repr(given_id))
    return valid_id


def add_all_species(model, species: List, initial_condition_dictionary: dict, compartment=None, **kwargs):
    """adds a list of Species to the SBML model.
    :param model: valid SBML model
    :param species: list of species to be added to the SBML model
    :param compartment: compartment id, if empty species go to the first compartment
    :param initial_concentration_dict: a dictionary s --> initial_concentration
    :return: None
    """


    elementlist = model.getSBMLDocument().getListOfAllElements()

    all_ids = getAllIds(elementlist)

    trans = SetIdFromNames(all_ids)

    all_ids = [str(s) for s in species]
    all_sbml_ids = [trans.getValidIdForName(species_id) for species_id in all_ids]
    



    for s_ind,s in enumerate( species):
        s_id = all_sbml_ids[s_ind]
        if compartment is None or s.compartment is not None:
            # If no compartment was passed in or if species (s) has its own compartment set:
            compartment = get_compartment_by_name(model, s.compartment.name)
            if compartment is None:
                compartment = add_compartment(model, s.compartment)
        if s in initial_condition_dictionary:
            initial_concentration = parameter_to_value(
                initial_condition_dictionary[s])
        else:
            initial_concentration = 0
        add_species(model=model, compartment=compartment, species_name=str(s.name), 
                            species_id=s_id, initial_concentration=initial_concentration)

def add_species(model, compartment, species_name, species_id, initial_concentration=None,  **kwargs):
    """Helper function to add a species to the sbml model.
    :param model:
    :param compartment: a compartment in the SBML model
    :param species: must be chemical_reaction_network.species objects
    :param initial_concentration: initial concentration of the species in the SBML model
    :return: SBML species object
    """

    # Construct the species ID

    logger.debug(f'Adding species: {species_name}, id: {species_id}')
    sbml_species = model.createSpecies()
    sbml_species.setName(species_name)
    sbml_species.setId(species_id)
    sbml_species.setCompartment(compartment.getId())
    sbml_species.setConstant(False)
    sbml_species.setBoundaryCondition(False)
    sbml_species.setHasOnlySubstanceUnits(False)
    sbml_species.setSubstanceUnits('mole')
    if initial_concentration is None:
        initial_concentration = 0
    sbml_species.setInitialConcentration(initial_concentration)

    return sbml_species

def add_all_compartments(model, compartments: List, **keywords):
    """ Adds the list of Compartment objects to the SBML model
    :param model: valid SBML model
    :param compartments: list of compartments to be added to the SBML model
    :return: None
    """
    for compartment in compartments:
        add_compartment(model = model, compartment = compartment, **keywords)

def add_compartment(model, compartment, **keywords):
    """ Helper function to add a compartment to the SBML model.
    :param model: a valid SBML model
    :param compartment: a Compartment object
    :return: SBML compartment object 
    """
    sbml_compartment = model.createCompartment()
    compartment_id = compartment.name
    sbml_compartment.setId(compartment_id)
    sbml_compartment.setName(compartment.name)
    sbml_compartment.setConstant(True)  # keep compartment size constant
    sbml_compartment.setSpatialDimensions(compartment.spatial_dimensions)  # For example, 3 dimensional compartment
    sbml_compartment.setSize(compartment.size)  # For example, 1e-6 liter
    if compartment.unit is not None:
        sbml_compartment.setUnits(compartment.unit)
    return sbml_compartment

def get_compartment_by_name(model, compartment_name):
    """ Helper function to find the SBML compartment object 
    given the compartment name in the SBML file
    """
    for compartment in model.getListOfCompartments():
        if compartment.getName() == compartment_name:
            return compartment

# Helper function to add a parameter to the model


def add_parameter(mixture, name, value=0, debug=False):
    model = mixture.model  # Get the model where we will store results

    # Check to see if this parameter is already present
    parameter = find_parameter(mixture, name)  # ! TODO: add error checking
    if parameter is None:
        parameter = model.createParameter()
        all_ids = getAllIds(model.getSBMLDocument().getListOfAllElements())
        trans = SetIdFromNames(all_ids)
        # ! TODO: add error checking
        parameter.setId(trans.getValidIdForName(name))
    # Set the value of the parameter
    parameter.setValue(float(value))
    parameter.setConstant(True)

    return parameter


# Look for a parameter in the current model
def find_parameter(mixture, id):
    model = mixture.model  # Get model where parameters are stored
    return model.getParameter(id)  # ! TODO: add error checking


def add_all_reactions(model, reactions: List, stochastic=False, **kwargs):
    """adds a list of reactions to the SBML model.
    :param model: an sbml model created by create_sbml_model()
    :param reactions: list of Reactions
    :param stochastic: binary flag for stochastic models
    :return: None
    """

    elementlist = model.getSBMLDocument().getListOfAllElements()

    all_ids = getAllIds(elementlist)

    trans = SetIdFromNames(all_ids)

    all_ids = [f"r{i}" for i in range(len(reactions))]
    all_ids_rev = [f"r{i}rev" for i in range(len(reactions))]
    all_sbml_ids = [trans.getValidIdForName(reaction_id) for reaction_id in all_ids]
    all_sbml_ids_rev = [trans.getValidIdForName(reaction_id) for reaction_id in all_ids_rev]

    for rxn_count, r in enumerate(reactions):

        add_reaction(model=model, crn_reaction=r,
                     reaction_id=all_sbml_ids[rxn_count], stochastic=stochastic, **kwargs)
        # Reversible reactions are always seperated into two seperate reactions
        if r.is_reversible:
            add_reaction(model=model, crn_reaction=r, reaction_id=all_sbml_ids_rev[rxn_count],
                         stochastic=stochastic, reverse_reaction=True, **kwargs)
def add_reaction(model, crn_reaction, reaction_id: str, stochastic: bool = False, reverse_reaction: bool = False, **kwargs):
    """adds a sbml_reaction to an sbml model.
    :param model: an sbml model created by create_sbml_model()
    :param crn_reaction: must be a chemical_reaction_network.reaction object
    :param reaction_id: unique id of the reaction
    :param stochastic: stochastic model flag
    :param reverse_reaction: 
    :return: SBML Reaction object
    """
    # Create the sbml_reaction in SBML
    sbml_reaction = model.createReaction()
    
    sbml_reaction.setId(reaction_id)
    sbml_reaction.setName(sbml_reaction.getId())
    # all reactions are set to be non-reversible in BioCRNpyler because this is correct in deterministic and stochastic simulation.
    sbml_reaction.setReversible(False)
    # Create the reactants and products for the sbml_reaction
    if not reverse_reaction:
        _create_reactants(reactant_list=crn_reaction.inputs,
                          sbml_reaction=sbml_reaction, model=model)
        _create_products(product_list=crn_reaction.outputs,
                         sbml_reaction=sbml_reaction, model=model)
    else:
        _create_reactants(reactant_list=crn_reaction.outputs,
                          sbml_reaction=sbml_reaction, model=model)
        _create_products(product_list=crn_reaction.inputs,
                         sbml_reaction=sbml_reaction, model=model)
    
    # Create the kinetic law and corresponding local propensity parameters
    crn_reaction.propensity_type.create_kinetic_law(model=model,
                                                    sbml_reaction=sbml_reaction,
                                                    stochastic=stochastic,
                                                    crn_reaction=crn_reaction,
                                                    reverse_reaction=reverse_reaction,
                                                    **kwargs)
    
    # Create SpeciesModifierReference in SBML for species that are referred by the
    # KineticLaw but not in reactants or products
    _create_modifiers(crn_reaction=crn_reaction,
                      sbml_reaction=sbml_reaction, model=model)
    


    return sbml_reaction


def _create_reactants(reactant_list, sbml_reaction, model):
    for input in reactant_list:
        # What to do when there are multiple species with same name?
        species_id = str(input.species)
        reactant = sbml_reaction.createReactant()
        reactant.setSpecies(species_id)
        reactant.setConstant(False)
        reactant.setStoichiometry(input.stoichiometry)


def _create_products(product_list, sbml_reaction, model):
    for output in product_list:
        species_id = str(output.species)
        product = sbml_reaction.createProduct()
        product.setSpecies(species_id)
        product.setStoichiometry(output.stoichiometry)
        product.setConstant(False)


def _create_modifiers(crn_reaction, sbml_reaction, model):
    reactants_list = [str(i.species) for i in crn_reaction.inputs]
    products_list = [str(i.species) for i in crn_reaction.outputs]
    modifier_species = [
        str(i) for i in crn_reaction.propensity_type.propensity_dict['species'].values()]
    for modifier_id in modifier_species:
        modifier_id = str(modifier_id)
        if modifier_id not in reactants_list and modifier_id not in products_list:
            modifier = sbml_reaction.createModifier()
            modifier.setSpecies(modifier_id)

# Creates a local parameter SBML kinetic rate law


def _create_local_parameter(ratelaw, name, value, constant=True):
    param = ratelaw.createParameter()
    param.setId(name)
    param.setConstant(constant)
    param.setValue(value)
    return param

# Creates a global parameter SBML model


def _create_global_parameter(model, name, value, p_unit=None, constant=True):
    if p_unit is not None:
        if not isinstance(p_unit, str):
            raise ValueError(
                "Units for a parameter must be passed as strings.")
        unit_added = False
        for unit_definition in model.getListOfUnitDefinitions():
            if unit_definition.getId() == p_unit:
                unit_added = True
        if not unit_added:
            unit_created = create_new_unit_definition(model, p_unit)
            model.addUnitDefinition(unit_created)

    if model.getParameter(name) is None:
        param = model.createParameter()
        param.setId(name)
        param.setConstant(constant)
        param.setValue(value)
        if p_unit is not None:
            param.setUnits(str(p_unit))
    else:
        param = model.getParameter(name)
        if p_unit is not None:
            param.setUnits(str(p_unit))
    return param

##
# @file    setIdFromNames.py
# @brief   Utility program, renaming all SIds that also has
# names specified. The new id will be derived from
# the name, with all invalid characters removed.
##
# @author  Frank T. Bergmann
##
##
# <!--------------------------------------------------------------------------
# This sample program is distributed under a different license than the rest
# of libSBML.  This program uses the open-source MIT license, as follows:
##
# Copyright (c) 2013-2017 by the California Institute of Technology
# (California, USA), the European Bioinformatics Institute (EMBL-EBI, UK)
# and the University of Heidelberg (Germany), with support from the National
# Institutes of Health (USA) under grant R01GM070923.  All rights reserved.
##
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
##
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
##
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
##
# Neither the name of the California Institute of Technology (Caltech), nor
# of the European Bioinformatics Institute (EMBL-EBI), nor of the University
# of Heidelberg, nor the names of any contributors, may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# ------------------------------------------------------------------------ -->
##
##

# This class implements an identifier transformer, that means it can be used
# to rename all sbase elements.


class SetIdFromNames(libsbml.IdentifierTransformer):
    def __init__(self, ids):
        # call the constructor of the base class
        libsbml.IdentifierTransformer.__init__(self)
        # remember existing ids ...
        self.existingIds = ids

        # The function actually doing the transforming. This function is called

    # once for each SBase element in the model.
    def transform(self, element):
        # return in case we don't have a valid element
        if (element is None
                or element.getTypeCode() == libsbml.SBML_LOCAL_PARAMETER):
            return libsbml.LIBSBML_OPERATION_SUCCESS

            # or if there is nothing to do
        if (element.isSetName() == False
                or element.getId() == element.getName()):
            return libsbml.LIBSBML_OPERATION_SUCCESS

            # find the new id
        newId = self.getValidIdForName(element.getName())

        # set it
        element.setId(newId)

        # remember it
        self.existingIds.append(newId)

        return libsbml.LIBSBML_OPERATION_SUCCESS

    def nameToSbmlId(self, name):
        IdStream = []
        count = 0
        end = len(name)

        if '0' <= name[count] and name[count] <= '9':
            IdStream.append('x_')
        if '*' in name:
            IdStream.append('xx')
        for count in range(0, end):
            if (('0' <= name[count] and name[count] <= '9') or
                    ('a' <= name[count] and name[count] <= 'z') or
                    ('A' <= name[count] and name[count] <= 'Z')):
                IdStream.append(name[count])
            else:
                IdStream.append('_')
        Id = ''.join(IdStream)
        if (Id[len(Id) - 1] != '_'):
            return Id

        return Id
        # return Id[:-1] #this code was removing underscores at the end of ComplexSpecies needed for imbedded ComplexSpecies.

    #
    # Generates the id out of the name, and ensures it is unique.
    # It does so by appending numbers to the original name.
    #
    def getValidIdForName(self, name):
        baseString = self.nameToSbmlId(name)
        id = baseString
        count = 1
        while (self.existingIds.count(id) != 0):
            id = "{0}_{1}".format(baseString, count)
            count = count + 1
        return id

    #  #


#  # Returns a list of all ids from the given list of elements
#  #
def getAllIds(allElements):
    result = []
    if allElements is None or allElements.getSize() == 0:
        return result

    for i in range(0, allElements.getSize()):
        current = allElements.get(i)
        if (current.isSetId()
                and current.getTypeCode() != libsbml.SBML_LOCAL_PARAMETER):
            result.append(current.getId())
    return result


def getSpeciesByName(model, name, compartment=''):
    '''
    Returns a list of species in the Model with the given name
    compartment : (Optional) argument to specify the compartment name in which
    to look for the species.
    '''
    if type(name) is not str:
        raise ValueError(
            f'"name" must be a string. Received {name}.')
    species_found = []
    for species in model.getListOfSpecies():
        if species.getName() == name:
            if compartment != '':
                comp_elem = species.getCompartment()
                comp_name = model.getElementBySId(comp_elem).getName()
                if comp_name == compartment:
                    species_found.append(species)
                else:
                    continue
            else:
                species_found.append(species)

    if len(species_found) == 1:
        return species_found[0]
    elif not species_found:
        raise ValueError('The species ' + name + ' not found.')
    else:
        warn('Multiple species with name ' + name + ' found. Returning a list')
        return species_found


# Validate SBML


class validateSBML(object):
    '''
    libSBML class to validate the generated SBML models
    ## @brief   Validates SBMLDocument
    ## @author  Akiya Jouraku (translated from libSBML C++ examples)
    ## @author  Ben Bornstein
    ## @author  Michael Hucka
    '''

    def __init__(self, ucheck):
        self.reader = libsbml.SBMLReader()
        self.ucheck = ucheck

    def validate(self, sbml_document, print_results=False):
        """sbml_document: libSBML SBMLDocument object.
        print_results: Print toggle for validation warnings.
        """
        sbmlDoc = sbml_document
        errors = sbmlDoc.getNumErrors()
        if print_results:
            print("Validating SBML model with ID: {0}...".format(
                sbmlDoc.getModel().getId()))
        seriousErrors = False

        numReadErr = 0
        numReadWarn = 0
        errMsgRead = ""

        if errors > 0:
            for i in range(errors):
                severity = sbmlDoc.getError(i).getSeverity()
                if (severity == libsbml.LIBSBML_SEV_ERROR) or (severity == libsbml.LIBSBML_SEV_FATAL):
                    seriousErrors = True
                    numReadErr += 1
                else:
                    numReadWarn += 1
            errMsgRead = sbmlDoc.getErrorLog().toString()

        # If serious errors are encountered while reading an SBML document, it
        # does not make sense to go on and do full consistency checking because
        # the model may be nonsense in the first place.

        numCCErr = 0
        numCCWarn = 0
        errMsgCC = ""

        if seriousErrors:
            errMsgRead += "Further consistency checking and validation aborted."
        else:
            sbmlDoc.setConsistencyChecks(
                libsbml.LIBSBML_CAT_UNITS_CONSISTENCY, self.ucheck)
            failures = sbmlDoc.checkConsistency()
            if failures > 0:
                isinvalid = False
                for i in range(failures):
                    severity = sbmlDoc.getError(i).getSeverity()
                    if (severity == libsbml.LIBSBML_SEV_ERROR) or (severity == libsbml.LIBSBML_SEV_FATAL):
                        numCCErr += 1
                        isinvalid = True
                    else:
                        numCCWarn += 1
                if isinvalid:
                    errMsgCC = sbmlDoc.getErrorLog().toString()
        if errMsgRead or errMsgCC:
            if print_results:
                print()
                print("===== validation error/warning messages =====\n")
            if errMsgRead:
                if print_results:
                    print(errMsgRead)
            if errMsgCC:
                if print_results:
                    print("*** consistency check ***\n")
                    print(errMsgCC)
        if not (numReadErr + numCCErr):
            print('Successful!')
        return numReadErr + numCCErr


def validate_sbml(sbml_document, enable_unit_check=False, print_results=True):
    """Validates the generated SBML model by using libSBML SBML validation code."""
    validator = validateSBML(enable_unit_check)
    validation_result = validator.validate(
        sbml_document, print_results=print_results)
    if validation_result > 0:
        warn('SBML model invalid. Run with print_results = False to hide print statements')
    return validation_result




#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

import copy
import warnings
from typing import Dict, List, Tuple, Union
from warnings import warn
import numbers

import libsbml

from .reaction import Reaction
from .sbmlutil import add_all_reactions, add_all_species, add_all_compartments, create_sbml_model
from .species import Species
from .utils import process_initial_concentration_dict, parameter_to_value, remove_bindloc
from .parameter import ModelParameter, Parameter


class ChemicalReactionNetwork(object):
    """A chemical reaction network is a container of species and reactions
    chemical reaction networks can be compiled into SBML.

    reaction types:
    mass action: standard mass action semantics where the propensity of a
    reaction is given by deterministic propensity =
    .. math::
    k \Prod_{inputs i} [S_i]^a_i
    stochastic propensity =
    .. math::
    k \Prod_{inputs i} (S_i)!/(S_i - a_i)!
    where a_i is the spectrometric coefficient of species i
    """
    def __init__(self, species: List[Species], reactions: List[Reaction],  initial_concentration_dict: Dict[Species,Union[numbers.Real, Parameter]] = None, show_warnings=False):
        self.species = species
        self.reactions = reactions
        self.initial_concentration_dict = None #Create an unpopulated dictionary
        self.initial_concentration_dict = initial_concentration_dict #update it

        ChemicalReactionNetwork.check_crn_validity(self._reactions, self._species, show_warnings=show_warnings)

    @property
    def species(self):
        return copy.deepcopy(self._species)

    @species.setter
    def species(self, species):
        if not hasattr(self, "_species"):
            self._species = []
            self._species_dict = {}
            self.add_species(species)
        else:
            raise AttributeError("The species in a CRN cannot be removed or modified. New Species can be added with CRN.add_species(...).")

    @property
    def reactions(self):
        return  copy.deepcopy(self._reactions)

    @reactions.setter
    def reactions(self, reactions):
        if not hasattr(self, "_reactions"):
            self._reactions = []
            self.add_reactions(reactions)
        else:
            raise AttributeError("The reactions in a CRN cannot be removed or modified. New reactions can be added with CRN.add_reactions(...).")

    
    def add_species(self, species, copy_species = True):
        """Adds a Species or a list of Species to the CRN object

        :param species: Species instance or list of Species instances
        :param copy_species: whether to deep copy Species added to the CRN. Protects CRN validity at teh expense of speed.

        """
        if not isinstance(species, list):
            species = [species]

        species = Species.flatten_list(species) #Flatten the list
        species = remove_bindloc(species)

        #Deepcopy the specied list here, to preserve its structure
        if copy_species:
            species = copy.deepcopy(species)

        for s in species:
            if not isinstance(s, Species): #check species are Species
                raise ValueError("A non-species object was used as a species!")
            if s not in self._species_dict: #Do not add duplicate Species
                self._species_dict[s] = True
                self._species.append(s) #copy the species and add it to the CRN

    def add_reactions(self, reactions: Union[Reaction,List[Reaction]], copy_reactions = True, add_species = True) -> None:
        """Adds a reaction or a list of reactions to the CRN object

        :param reactions: Reaction instance or list of Reaction instances
        :param copy_reactions: whether to deep copy reactions before adding them to the CRN. Protects CRN validity at the expense of speed.
        :param add_species: whether to add species in reactions to the CRN. Prevents errors at the expense of speed.
        :return: None
        """
        if not isinstance(reactions, list):
            reactions = [reactions]

        #It is recommended to copy reactions before adding them to the CRN, so they are "protected"
        if copy_reactions:
            reactions = copy.deepcopy(reactions) #deep copy all the reactions

        #Add the reactions to the CRN
        self._reactions += reactions

        #Add species from reactions into the CRN
        if add_species:
            for r in reactions:
                if not isinstance(r, Reaction): # check reactions and Reactions
                    raise ValueError("A non-reaction object was used as a reaction!")

                # add all the Species in the reaction to the CRN
                reaction_species = list(set([w.species for w in r.inputs + r.outputs]))
                self.add_species(reaction_species, copy_species = copy_reactions)

    @property
    def initial_concentration_dict(self):
        return self._initial_concentration_dict

    @initial_concentration_dict.setter
    def initial_concentration_dict(self, initial_concentration_dict):
        if initial_concentration_dict is None:
            self._initial_concentration_dict = {}
        elif isinstance(initial_concentration_dict, dict):
            for s in initial_concentration_dict:
                if s not in self._species_dict:
                    raise ValueError(f"Trying to set the initial concentration of a Species {s} not in the CRN")
                elif parameter_to_value(initial_concentration_dict[s]) >= 0:
                    self.initial_concentration_dict[s] = initial_concentration_dict[s]
                else:
                    raise ValueError(f"Trying to set a species {s} to a negative concentration {initial_concentration_dict[s]}")

    @staticmethod
    def check_crn_validity(reactions: List[Reaction], species: List[Species], show_warnings=True) -> Tuple[List[Reaction],List[Species]]:
        """Checks that the given list of reactions and list of species can form a valid CRN.

        :param reactions: list of reaction
        :param species: list of species
        :param show_warnings: whether to show warning when duplicated reactions/species was found
        :return: tuple(reaction,species)
        """

        if not all(isinstance(r, Reaction) for r in reactions):
            raise ValueError("A non-reaction object was used as a reaction!")

        if not all(isinstance(s, Species) for s in species):
            raise ValueError("A non-species object was used as a species!")

        for r in reactions:
            if reactions.count(r) > 1 and show_warnings:
                warn(f"Reaction {r} may be duplicated in CRN definitions. "
                     f"Duplicates have NOT been removed.")

        for s in species:
            if species.count(s) > 1 and show_warnings:
                warn(f"Species {s} is duplicated in the CRN definition. "
                     f"Duplicates have NOT been removed.")

        # check that all species in the reactions are also in the species list and vice versa
        unique_species = set(species)
        all_species_in_reactions = set(Species.flatten_list([r.species for r in reactions]))
        if unique_species != all_species_in_reactions:
            species_without_reactions = unique_species - all_species_in_reactions
            if species_without_reactions and show_warnings:
                warn(f'These Species {list(species_without_reactions)} are not part of any reactions in the CRN!')
            unlisted_reactions = all_species_in_reactions - unique_species
            if unlisted_reactions and show_warnings:
                warn(f'These Species {list(unlisted_reactions)} are not listed in the Species list, but part of the reactions!')

        return reactions, species

    def __repr__(self):
        txt = "Species = "
        for s in self._species:
            txt += repr(s) + ", "
        txt = txt[:-2] + '\n'
        txt += "Reactions = [\n"

        for r in self._reactions:
            txt += "\t" + repr(r) + "\n"
        txt += "]"
        return txt

    def pretty_print(self, show_rates = True, show_material = True, show_attributes = True, show_compartment = False,
                        show_initial_concentration = True, show_keys = True, **kwargs):
        """A more powerful printing function.

        Useful for understanding CRNs but does not return string identifiers.
        show_material toggles whether species.material is printed.
        show_attributes toggles whether species.attributes is printed
        show_rates toggles whether reaction rate functions are printed
        """

        txt = "Species"+ f"(N = {len(self._species)}) = "+"{\n"
        
        ics = lambda s: self.initial_concentration_dict[s] if s in self.initial_concentration_dict else 0

        species_sort_list = [(parameter_to_value(ics(s)), s) for s in self._species]
        species_sort_list.sort()
        species_sort_list.reverse()
        for sind, (init_conc, s) in enumerate(species_sort_list):
            init_conc = ics(s) 

            txt += s.pretty_print(show_material = show_material, show_compartment = False, show_attributes = show_attributes, **kwargs)

            if show_initial_concentration:
                txt += f" (@ {parameter_to_value(init_conc)}),  "

                if show_keys: #shows where the initial conditions came from
                    if isinstance(init_conc, ModelParameter):
                        txt+=f"\n   found_key=(mech={init_conc.found_key.mechanism}, partid={init_conc.found_key.part_id}, name={init_conc.found_key.name}).\n   search_key=(mech={init_conc.search_key.mechanism}, partid={init_conc.search_key.part_id}, name={init_conc.search_key.name}).\n"


        txt += '\n}\n'
        txt += f"\nReactions ({len(self._reactions)}) = [\n"

        for rind in range(len(self._reactions)):
            r = self._reactions[rind]
            txt += f"{rind}. " + r.pretty_print(show_rates = show_rates, show_material = show_material, show_attributes = show_attributes, show_keys = show_keys, **kwargs) + "\n"
        txt += "]"
        return txt

    def initial_condition_vector(self, init_cond_dict: Union[Dict[str, float], Dict[Species, float]]):
        x0 = [0.0] * len(self._species)
        for idx, s in enumerate(self._species):
            if s in init_cond_dict:
                x0[idx] = init_cond_dict[s]
        return x0

    def get_all_species_containing(self, species: Species, return_as_strings = False):
        """Returns all species (complexes and otherwise) containing a given species
           (or string).
        """
        return_list = []
        if not isinstance(species, Species):
            raise ValueError('species argument must be an instance of Species!')

        for s in self._species:
            if species in s.get_species(recursive = True):
                if return_as_strings:
                    return_list.append(repr(s))
                else:
                    return_list.append(s)
        return return_list

    def replace_species(self, species: Species, new_species: Species):
        """Replaces species with new_species in the entire CRN.

        Does not act in place: returns a new CRN.
        """

        if not isinstance(species, Species):
            raise ValueError('species argument must be an instance of Species!')

        if not isinstance(new_species, Species):
            raise ValueError('species argument must be an instance of Species!')

        new_species_list = []
        for s in self._species:
            new_s = s.replace_species(species, new_species)
            new_species_list.append(new_s)

        new_reaction_list = []
        for r in self._reactions:
            new_r = r.replace_species(species, new_species)
            new_reaction_list.append(new_r)

        return ChemicalReactionNetwork(new_species_list, new_reaction_list)

    def generate_sbml_model(self, stochastic_model=False, show_warnings = False, check_validity = True,**keywords):
        """Creates an new SBML model and populates with the species and
        reactions in the ChemicalReactionNetwork object

        :param stochastic_model: whether the model is stochastic
        :param show_warnings: of from check crn validity
        :param keywords: extra keywords pass onto create_sbml_model() and add_all_reactions()
        :return: tuple: (document,model) SBML objects
        """
        if(check_validity):
            ChemicalReactionNetwork.check_crn_validity(self._reactions, self._species, show_warnings=show_warnings)
        
        document, model = create_sbml_model(**keywords)
        all_compartments = []
        for species in self._species:
            if species.compartment not in all_compartments: 
                all_compartments.append(species.compartment)
        add_all_compartments(model = model, compartments = all_compartments, **keywords)
        
        add_all_species(model=model, species=self._species, initial_condition_dictionary = self.initial_concentration_dict)
        add_all_reactions(model=model, reactions=self._reactions, stochastic_model=stochastic_model, **keywords)
        

        if document.getNumErrors():
            warn('SBML model generated has errors. Use document.getErrorLog() to print all errors.')
        return document, model

    def write_sbml_file(self, file_name=None, stochastic_model = False, check_validity = True, **keywords) -> bool:
        """"Writes CRN object to a SBML file

        :param file_name: name of the file where the SBML model gets written
        :param stochastic_model: export an SBML file which ready for stochastic simulations
        :param keywords: keywords that passed into generate_sbml_model()
        :return: bool, show whether the writing process was successful
        """
        document, _ = self.generate_sbml_model(stochastic_model = stochastic_model, check_validity = check_validity,**keywords)
        sbml_string = libsbml.writeSBMLToString(document)
        with open(file_name, 'w') as f:
            f.write(sbml_string)
        return True

    def simulate_with_bioscrape(self, timepoints, initial_condition_dict=None,
                                stochastic = False, return_dataframe = True,
                                safe = False):

        """Simulate CRN model with bioscrape (https://github.com/biocircuits/bioscrape).
        Returns the data for all species as Pandas dataframe.
        """
        result = None
        warnings.warn("simulate_with_bioscrape is depricated and will cease working in a future release. Instead, please use simulate_with_bioscrape_via_sbml.")
        
        result = self.simulate_with_bioscrape_via_sbml(timepoints, filename = None, 
            initial_condition_dict = initial_condition_dict, return_dataframe = return_dataframe, 
            safe = safe, stochastic = stochastic)

        return result

    def simulate_with_bioscrape_via_sbml(self, timepoints, filename = None,
                initial_condition_dict = None, return_dataframe = True,
                stochastic = False, safe = False, return_model = False, check_validity=True, **kwargs):

        """Simulate CRN model with bioscrape via writing a SBML file temporarily.
        [Bioscrape on GitHub](https://github.com/biocircuits/bioscrape).

        Returns the data for all species as Pandas dataframe.
        """
        result = None
        m = None
        try:
            from bioscrape.simulator import py_simulate_model
            from bioscrape.types import Model

            if filename is None:
                self.write_sbml_file(file_name ="temp_sbml_file.xml", stochastic_model = stochastic, for_bioscrape = True,check_validity=check_validity)
                file_name = "temp_sbml_file.xml"
            elif isinstance(filename, str):
                file_name = filename
            else:
                raise ValueError(f"filename must be None or a string. Recievied: {filename}")

            if 'sbml_warnings' in kwargs:
                sbml_warnings = kwargs.get('sbml_warnings')
            else:
                sbml_warnings = False
            m = Model(sbml_filename = file_name, sbml_warnings = sbml_warnings)
            # m.write_bioscrape_xml('temp_bs'+ file_name + '.xml') # Uncomment if you want a bioscrape XML written as well.
            if initial_condition_dict is not None:
                processed = process_initial_concentration_dict(initial_condition_dict)
                m.set_species(processed)
            result = py_simulate_model(timepoints, Model = m, stochastic = stochastic, safe = safe,
                                                return_dataframe = return_dataframe)
        except ModuleNotFoundError:
            warnings.warn('bioscrape was not found, please install bioscrape')

        if return_model:
            return result, m
        else:
            return result

    def simulate_with_roadrunner(self, timepoints: List[float], initial_condition_dict: Dict[str,float]=None, return_roadrunner=False, check_validity=True):
        """To simulate using roadrunner.
        Arguments:
        timepoints: The array of time points to run the simulation for.
        initial_condition_dict:

        Returns the results array as returned by RoadRunner OR a Roadrunner model object.

        Refer to the libRoadRunner simulator library documentation 
        for details on simulation results: (http://libroadrunner.org/)[http://libroadrunner.org/]
        NOTE : Needs roadrunner package installed to simulate.
        """
        res_ar = None
        try:
            import roadrunner
            import io
            document, _ = self.generate_sbml_model(stochastic_model=False, check_validity=check_validity)
            sbml_string = libsbml.writeSBMLToString(document)
            # write the sbml_string into a temporary file in memory instead of a file
            string_out = io.StringIO()
            string_out.write(sbml_string)
            # use the temporary file in memory to load the model into libroadrunner
            rr = roadrunner.RoadRunner(string_out.getvalue())
            if initial_condition_dict:
                for species, value in initial_condition_dict.items():
                    rr.model[f"init([{species}])"] = value

            if return_roadrunner:
                return rr
            else:
                result = rr.simulate(timepoints[0], timepoints[-1], len(timepoints))
                res_ar = result
        except ModuleNotFoundError:
            warnings.warn('libroadrunner was not found, please install libroadrunner')
        return res_ar




#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.from .dna_part import DNA_part

from .dna_part_misc import DNA_part


class Terminator(DNA_part):
    def __init__(self,name, **keywords):
        DNA_part.__init__(self,name, **keywords)
        self.name = name
    def update_species(self):
        return []
    def update_reactions(self):
        return []



# parameter.py - parameter processing
# RMM, 19 Aug 2018
#
# This file contains the Parameter class that is used for representing
# parameters, as well as utility functions for manipulating
# parameters.
#
# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.


"""
#### Parameter Value Defaulting:
Not all parameters need to have the required headings. 
    The only two required columns are "param_val" and "param_name". 
    BioCRNpyler uses a form of parameter name defaulting discussed below to find default parameters 
    if no exact match is in the config file. This makes it easy to set default parameters for things 
    like "ku" and "ktx" to quickly build models.

#### Parameters inside BioCRNpyler:
Inside of bioCRNpyler, parameters are stored as a dictionary key value pair: 
    (mechanism_name, part_id, param_name) --> param_val. If that particular parameter key cannot be found, 
    the software will default to the following keys: 
    (mechanism_type, part_id, param_name) >> (part_id, param_name) >> 
    (mechanism_name, param_name) >> (mechanism_type, param_name) >>
       (param_name) and give a warning. 
    As a note, mechanism_name refers to the .name variable of a Mechanism. mechanism_type refers to the .type variable of a Mechanism. 
    Either of these can be used as a mechanism_id. This allows for models to be constructed easily using default parameter values and 
    for parameters to be shared between different Mechanisms and/or Components.

#### Initial Conditions are also Parameters
The initial condition of any Species (or Component) will also be looked up as a parameters automatically.
    Initial conditions can be customized in through the custom_initial_conditions keyword in the Mixture constructor.
    custom_initial_conditions will take precedent to parameter initial conditions.

#### Units are read directly read from the column labeled "units" in the parameter file. 

"""

import csv
import numbers
import re
from collections import namedtuple  # Used for the parameter keys
from typing import Dict, List, Union
from warnings import warn

ParameterKey = namedtuple('ParameterKey', 'mechanism part_id name')  # This could later be extended


class Parameter(object):
    def __init__(self, parameter_name: str, parameter_value: Union[str, numbers.Real], unit=None):
        """A class for representing parameters in general. Only the below subclasses are ever used.

        :param parameter_name: is the name of the parameter
        :param parameter_value: is the value of the parameter
        :param unit: is the unit of the parameter or a species
        """
        self.parameter_name = parameter_name
        self.value = parameter_value
        self.unit = unit

    @property
    def parameter_name(self) -> str:
        return self._parameter_name

    @parameter_name.setter
    def parameter_name(self, new_parameter_name: str):
        if not isinstance(new_parameter_name, str):
            raise ValueError(f"parameter_name must be a string: received {type(new_parameter_name)}.")
        if not re.search('^[a-z]+', new_parameter_name, re.IGNORECASE):
            raise ValueError(f'parameter_name should be at least one character and cannot start with a number!')

        self._parameter_name = new_parameter_name

    @property
    def value(self) -> numbers.Real:
        return self._value

    @value.setter
    def value(self, new_parameter_value: Union[str, numbers.Real]):
        if not (isinstance(new_parameter_value, numbers.Real) or isinstance(new_parameter_value, str)):
            raise ValueError(f"parameter_value must be a float or int: received {type(new_parameter_value)}.")
        if isinstance(new_parameter_value, str):
            if re.search('[a-df-z]', new_parameter_value, re.I) \
                    or re.search('(^[1-9]+/[1-9]+)|(^[1-9]+e-?[0-9]+)|(^.?[0-9])', new_parameter_value, re.I) is None:
                raise ValueError(f'No valid parameter value! Accepted formats: 1.00 or 1e4 or 2/5, we got {new_parameter_value} ')

            self._value = Parameter._convert_rational(new_parameter_value)
        else:
            self._value = new_parameter_value

    @property
    def unit(self) -> str:
        return self._unit
    
    @unit.setter
    def unit(self, new_unit: str):
        if new_unit is None:
            self._unit = ""
        elif type(new_unit) is not str:
            raise ValueError(f"All units must be strings. Recieved {new_unit}.")
        else:
            self._unit = new_unit

    @staticmethod
    def _convert_rational(p_value: str) -> numbers.Real:
        if '/' in p_value:
            nom, denom = p_value.split('/')
            return float(nom)/float(denom)
        else:
            return float(p_value)
    def __eq__(self, other):
        if isinstance(other,Parameter):
            return self.value == other.value
        else:
            try:
                return self.value == float(other)
            except TypeError:
                raise TypeError(f"Cannot compare parameter {self} with {other}.")
        
    def __str__(self):
        return f"Parameter {self.parameter_name} = {self.value}"
    def __hash__(self):
        return hash(self._parameter_name)+hash(self._value)+hash(self._unit)


class ParameterEntry(Parameter):
    """A class for representing parameters in a parameter stored the ParameterDatabase.

     parameter_keys is a dictionary {key:value} or named_tuple (type ParameterKey) of keys for looking up the parameter
     parameter_info is a dictionary {key:value} of additional information about the parameter. 
         For example: additional columns in the parameter file or the parameter file name.
    """
    def __init__(self, parameter_name: str, parameter_value: Union[str,numbers.Real], parameter_key=None, parameter_info=None, **keywords):
        Parameter.__init__(self, parameter_name, parameter_value, **keywords)

        self.parameter_key = parameter_key
        self.parameter_info = parameter_info

    # Helper function to create ParameterKeys
    @staticmethod
    def create_parameter_key(new_key: Union[Dict, ParameterKey, str], parameter_name=None) -> ParameterKey:
        # New Key can be a named_tuple
        if isinstance(new_key, dict):
            new_key = dict(new_key)
            if parameter_name is not None:
                new_key["name"] = parameter_name
            for k in ParameterKey._fields:
                if k not in new_key:
                    new_key[k] = None
            return ParameterKey(**new_key) #automatically unpack the keywords
        elif isinstance(new_key, ParameterKey):
            return new_key
        elif isinstance(new_key, tuple) and len(list(new_key)) == len(ParameterKey._fields):
            # make a dictionary assuming correct ordering
            keywords = {ParameterKey._fields[i]:new_key[i] for i in range(len(ParameterKey._fields))}
            return ParameterKey(**keywords) #automatically unpack the keywords
        elif isinstance(new_key, str):
            return ParameterKey(mechanism = None, part_id = None, name = new_key)
        elif new_key is None and parameter_name is not None:
            return ParameterKey(mechanism = None, part_id = None, name = parameter_name)
        else:
            raise ValueError(f"parameter_key must be None, a dictionary, a ParameterKey, a {len(ParameterKey._fields)}-tuple, or a string (parameter name): received {new_key}.")

    @property
    def parameter_key(self) -> ParameterKey:
        return self._parameter_key

    @parameter_key.setter
    def parameter_key(self, parameter_key: Union[Dict, ParameterKey, str]):
        self._parameter_key = self.create_parameter_key(parameter_key, self.parameter_name)

    @property
    def parameter_info(self) -> Dict:
        return self._parameter_info

    @parameter_info.setter
    def parameter_info(self, parameter_info: Dict):
        if parameter_info is None:
            self._parameter_info = {}
        elif isinstance(parameter_info, dict):
            self._parameter_info = dict(parameter_info)

            #Update the units attribute, if necessary
            if "unit" in parameter_info and self.unit != "" and self.unit != parameter_info["unit"]:
                raise ValueError(f"Recieved multiple parameter units through constructor {self.unit} and parameter_info dictionary {parameter_info['unit']}.")
            elif "unit" in parameter_info:
                self.unit = parameter_info["unit"]

        else:
            raise ValueError(f"parameter_info must be None or a dictionary: received {parameter_info}.")

    def get_sbml_id(self):
        sbml_id = self.parameter_key.name+"_"
        if self.parameter_key.part_id is not None:
            sbml_id += self.parameter_key.part_id
        sbml_id += "_"
        if self.parameter_key.mechanism is not None:
            sbml_id += self.parameter_key.mechanism
        return sbml_id

    def __str__(self):
        return f"ParameterEntry({self.parameter_key}) = {self.value}"


class ModelParameter(ParameterEntry):
    """A class for representing parameters used in the Model.

      search_key is a tuple searched for to find the parameter, eg (mech_id, part_id, param_name), :
      found_key is the tuple used after defaulting to find the parameter eg (param_name)
    """
    def __init__(self, parameter_name: str, parameter_value: Union[str, numbers.Real], search_key, found_key, unit=None, parameter_key=None, parameter_info=None, **keywords):

        ParameterEntry.__init__(self, parameter_name, parameter_value, unit=unit, parameter_key=parameter_key, parameter_info=parameter_info, **keywords)
        self.search_key = search_key
        self.found_key = found_key

    @property
    def search_key(self):
        return self._search_key

    @search_key.setter
    def search_key(self, search_key):
        self._search_key = self.create_parameter_key(search_key, self.parameter_name)

    @property
    def found_key(self):
        return self._found_key

    @found_key.setter
    def found_key(self, found_key):
        self._found_key = self.create_parameter_key(found_key, self.parameter_name)

    def __str__(self):
        return f"ModelParameter({self.parameter_key}) = {self.value}\tsearch_key={self.search_key}"


class ParameterDatabase(object):
    def __init__(self, parameter_dictionary=None, parameter_file=None, overwrite_parameters=False):
        """A class for storing parameters in Components and Mixtures.

        :param parameter_dictionary:
        :param parameter_file:
        :param overwrite_parameters: whether to overwrite existing entries in the parameter database
        """

        self.parameters = {} #create an emtpy dictionary to get parameters.

        if isinstance(parameter_file, str):
            self.load_parameters_from_file(parameter_file, overwrite_parameters = overwrite_parameters)
        elif isinstance(parameter_file, list):
            for p in parameter_file:
                if isinstance(p, str):
                    self.load_parameters_from_file(p, overwrite_parameters = overwrite_parameters)
                else:
                    raise ValueError("parameter_file must be a string or list of strings representing file names and paths.")
        elif parameter_file is not None:
            raise ValueError("parameter_file must be a string representing a file name and path.")
            
        if isinstance(parameter_dictionary, dict):
            self.load_parameters_from_dictionary(parameter_dictionary, overwrite_parameters = overwrite_parameters)
        elif parameter_dictionary is not None:
            raise ValueError("parameter_dictionary must be None or a dictionary!")

    # To check if a key or ParameterEntry is in a the ParameterDatabase
    def __contains__(self, val):
        if isinstance(val, ParameterEntry):
            key = val.parameter_key
            if key in self.parameters and self.parameters[key] == val:
                return True
            else:
                return False
        else:
            try:
                key = ParameterEntry.create_parameter_key(val)
                return key in self.parameters
            except ValueError:
                return False

    # Ability to loop through parameters eg
    # for entry in ParameterDatabase: ...
    def __iter__(self):
        self.keys = list(self.parameters.keys())
        self.current_key_ind = 0
        return self

    def __next__(self):
        if self.current_key_ind < len(self.keys):
            key = self.keys[self.current_key_ind]
            entry = self.parameters[key]
            self.current_key_ind += 1
            return entry
        else:
            raise StopIteration

    # Length method
    def __len__(self):
        return len(self.parameters)

    # Gets a parameter from the database
    # Only returns exact matches.
    def __getitem__(self, key):
        param_key = ParameterEntry.create_parameter_key(key)
        return self.parameters[param_key]

    # Sets a parameter in the databases - useful for quickly changing parameters, but add_parameter is recommended.
    def __setitem__(self, parameter_key, value):

        key = ParameterEntry.create_parameter_key(parameter_key)
        
        if isinstance(value, ParameterEntry):
            if key != value.parameter_key:
                raise ValueError(f"Parameter Key does not match: ParameterDatabase key {key} is not the same as ParameterEntry Key {value.parameter_key}.")
            self.parameters[key] = value
        else:
            self.add_parameter(key.name, value, parameter_key = key, parameter_origin = "Set Manually", overwrite_parameters = True)

    def __str__(self):
        txt = "ParameterDatabase:"
        param_txt = "\n".join([repr(p) for p in self.parameters])
        return txt+param_txt

    def add_parameter(self, parameter_name: str, parameter_value: Union[str,numbers.Real], parameter_origin = None, parameter_key = None, parameter_info = None, overwrite_parameters = False):
        """Adds a parameter to the database with appropriate metadata

        :param parameter_name: the name of the parameter
        :param parameter_value: the value of the parameter
        :param parameter_origin:
        :param parameter_key:
        :param parameter_info:
        :param overwrite_parameters: whether to overwrite existing entries in the parameter database
        :return:
        """

        # Put parameter origin into parameter_info
        if parameter_info is None:
            parameter_info = {}
        if "parameter origin" not in parameter_info:
            parameter_info["parameter origin"] = parameter_origin

        # Create ParameterEntry
        param = ParameterEntry(parameter_name, parameter_value, parameter_key = parameter_key, parameter_info = parameter_info)
        key = param.parameter_key
        
        # Update parameter dictionary
        if key in self.parameters and not overwrite_parameters:
            raise ValueError(f"Duplicate parameter detected. Parameter with key = {key} is already in the ParameterDatabase. To Overwrite existing parameters, use overwrite_parameters = True.")
        else:
            self.parameters[key] = param

    def load_parameters_from_dictionary(self, parameter_dictionary: Dict[ParameterKey, Union[str,numbers.Real]], overwrite_parameters=False) -> None:
        """Loads Parameters from a parameter dictionary.

        :param parameter_dictionary: Dictionary with keys ParameterKey types and values with real numbers
        :param overwrite_parameters: whether to overwrite existing entries in the parameter database
        """
        for k in parameter_dictionary:
            key = ParameterEntry.create_parameter_key(k)
            self.add_parameter(key.name, parameter_dictionary[k], parameter_key = {"part_id":key.part_id, "mechanism":key.mechanism}, parameter_origin = "parameter_dictionary", overwrite_parameters = overwrite_parameters)

    def load_parameters_from_database(self, parameter_database, overwrite_parameters=False) -> None:
        """Loads parameters from another ParameterDatabase.

        :param parameter_database: instance of another ParameterDatabase
        :param overwrite_parameters:  whether to overwrite existing entries in the parameter database
        """

        if not isinstance(parameter_database, ParameterDatabase):
            raise TypeError(f"paramater_database must be a ParamaterDatabase: recievied {parameter_database}.")

        for k in parameter_database:
            if k not in self.parameters or overwrite_parameters:
                self.parameters[k.parameter_key] = parameter_database[k.parameter_key]
            else:
                raise ValueError(f"Duplicate parameter detected. Parameter with key = {k} is already in the ParameterDatabase. To Overwrite existing parameters, use overwrite_parameters = True.")

    def load_parameters_from_file(self, filename: str, overwrite_parameters=False) -> None:
        """Loads parameters from a file to the ParameterDatabase.

        Parameter files must be tab-separated (.tsv or .txt) or comma-separated (.csv) files!
        :param filename: name of the file (with valid file path)
        :param overwrite_parameters: whether to overwrite existing entries in the parameter database
        """

        # Figure out the format of the parameter file from the file extension
        with open(filename) as f:
            file_type = filename.split(".")[-1]
            if file_type in ["tsv", "txt"]:
                delimiter = '\t'
            elif file_type in ["csv"]:
                delimiter = ","
            else:
                raise ValueError("Parameter files must be tab-seperated (.tsv or .txt) or comma-seperated (.csv) files.")

            csvreader = csv.DictReader(f, delimiter=delimiter)
            # Used for flexible column headings
            accepted_field_names = {
                'mechanism': ['mechanism', 'mechanism_id'],
                'param_name': ["parameter_name", "parameter", "param", "param_name"],
                'part_id': ['part_id', 'part'],
                'param_val': ["val", "value", "param_val", "parameter_value"],
                'unit':['unit', 'units']
            }

            field_names = self._get_field_names(csvreader.fieldnames, accepted_field_names)

            # Determine which columns are in the CSV
            if field_names['param_name'] is None:
                warn('No param_name column was found, could not load parameter!')
            if field_names['mechanism'] is None:
                no_mechism_column = True
            else:
                no_mechism_column = False

            if field_names['part_id'] is None:
                no_part_id_column = True
            else:
                no_part_id_column = False

            #Load all parameters
            for row in csvreader:
                param_value = row[field_names['param_val']]
                field_columns = [field_names['param_name'], field_names['part_id'], field_names['mechanism'], field_names['param_val']]
                parameter_info = {k:row[k] for k in row if k not in field_columns}
                # TODO test all these cases!

                # Case 1: No Param Name so skip the row
                if row[field_names['param_name']] is None or len(row[field_names['param_name']]) == 0:
                    pass

                # Case 2: Just a Param Name
                elif no_mechism_column and no_part_id_column:
                    param_name = row[field_names['param_name']]
                    self.add_parameter(param_name, param_value, parameter_origin = filename, 
                        parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)

                # Case 3: Part_id and Param Name
                elif no_mechism_column and no_part_id_column is False:
                    param_name = row[field_names['param_name']]
                    part_id = row[field_names['part_id']]

                    if part_id is not None and len(part_id) > 0:
                        self.add_parameter(param_name, param_value, parameter_key = {"part_id":part_id}, parameter_origin = filename, 
                            parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)
                    else:
                        self.add_parameter(param_name, param_value, parameter_origin = filename, 
                            parameter_info = parameter_info , overwrite_parameters = overwrite_parameters)

                # Case 4: mechanism and param name
                elif no_part_id_column and no_mechism_column is False:
                    mech_name = row[field_names['mechanism']]
                    param_name = row[field_names['param_name']]
                    if mech_name is not None and len(mech_name) > 0:
                        self.add_parameter(param_name, param_value, parameter_key = {"mechanism":mech_name}, parameter_origin = filename, 
                        parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)
                    else:
                        self.add_parameter(param_name, param_value, parameter_origin = filename, 
                            parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)

                # Case 5: mechanism, part_id, and param name
                else:
                    part_id = row[field_names['part_id']]
                    mech_name = row[field_names['mechanism']]
                    param_name = row[field_names['param_name']]
                    if part_id is not None and len(part_id) > 0 and mech_name is not None and len(mech_name) >0:
                        self.add_parameter(param_name, param_value, parameter_key = {"part_id":part_id, "mechanism":mech_name}, parameter_origin = filename, 
                            parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)
                    
                    elif part_id is not None and len(part_id) > 0:    
                        self.add_parameter(param_name, param_value, parameter_key = {"part_id":part_id}, parameter_origin = filename, 
                            parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)

                    elif mech_name is not None and len(mech_name) >0:
                        self.add_parameter(param_name, param_value, parameter_key = {"mechanism":mech_name}, parameter_origin = filename, 
                            parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)
                    else:
                        self.add_parameter(param_name, param_value, parameter_origin = filename, 
                            parameter_info = parameter_info, overwrite_parameters = overwrite_parameters)

    @staticmethod
    def _get_field_names(field_names: List[str], accepted_field_names: Dict[str, List[str]]) -> Dict[str, str]:
        """Searches through valid field names and finds the currently used one. It builds a dictionary of currently
            used field names.

        :param field_names: list of field names (columns) found in the csv file
        :param accepted_field_names: dictionary of possible field names and their valid aliases
        :return: dictionary of currently used field names (aliases)
        """
        if not isinstance(field_names, list):
            raise ValueError('field_names must be a list of strings')
        if isinstance(field_names, list) and len(field_names) == 0:
            raise ValueError('field_names cannot be empty list!')
        if not isinstance(accepted_field_names, dict):
            raise ValueError('accepted_field_names must be a dictionary')
        if isinstance(accepted_field_names, dict) and len(accepted_field_names) == 0:
            raise ValueError('accepted_field_names cannot be empty dictionary')

        return_field_names = dict.fromkeys(accepted_field_names.keys())
        for accepted_name in accepted_field_names:
            # try to find an possible accepted names in the field_names using a generator
            try:
                loc_gen = (idx for idx, name in enumerate(accepted_field_names[accepted_name]) if name in field_names)
                loc_idx = next(loc_gen)
            except StopIteration:
                # we have reached the end of the possible names
                return_field_names[accepted_name] = None
                warn(f"parameter file contains no {accepted_name} column! Please add a "
                     f"column named {accepted_field_names[accepted_name]}.")
            else:
                return_field_names[accepted_name] = accepted_field_names[accepted_name][loc_idx]

        return return_field_names

    def find_parameter(self, mechanism, part_id, param_name):
        """Searches the database for the best matching parameter. 
        
        Parameter defaulting hierarchy:
        (mechanism_name, part_id, param_name) --> param_val. If that particular parameter key cannot be found, 
        the software will default to the following keys: 
        (mechanism_type, part_id, param_name) >> (part_id, param_name) >> 
        (mechanism_name, param_name) >> (mechanism_type, param_name) >>
        (param_name) and give a warning. 
        As a note, mechanism_name refers to the .name variable of a Mechanism. mechanism_type refers to the .type variable of a Mechanism. 
        Either of these can be used as a mechanism_id. This allows for models to be constructed easily using default parameter values and 
        for parameters to be shared between different Mechanisms and/or Components.
        """

        #this is imported here because otherwise there are import loops
        from .mechanism import Mechanism

        found_entry = None

        if isinstance(mechanism, str):
            mech_name = mechanism
            mech_type = mechanism
        elif isinstance(mechanism, Mechanism):
            mech_name = mechanism.name
            mech_type = mechanism.mechanism_type
        elif mechanism is not None:
            raise ValueError(f"mechanism keyword must be or string or have name and mechanism_type attributes: recievied {mechanism}.")
        else:
            mech_name = None
            mech_type = None
        
        parameter_key_list = [
            ParameterKey(mechanism = mech_name, part_id = part_id, name = param_name), 
            ParameterKey(mechanism = mech_type, part_id = part_id, name = param_name),
            ParameterKey(mechanism = None, part_id = part_id, name = param_name),
            ParameterKey(mechanism = mech_name, part_id = None, name = param_name),
            ParameterKey(mechanism = mech_type, part_id = None, name = param_name),
            ParameterKey(mechanism = None, part_id = None, name = param_name),
            ]

        for key in parameter_key_list:
            if key in self.parameters and found_entry is None:
                found_entry = self.parameters[key]
                found_key = key
                break

        if found_entry is None:
            return None
        else:
            return_param = ModelParameter(found_entry.parameter_name,found_entry.value, (mech_name, part_id, param_name), found_key,
                parameter_key = found_entry.parameter_key, parameter_info = found_entry.parameter_info, unit = found_entry.unit)
            return return_param



################################################################
#       DNA_construct: a higher level for construct compilation
#       Author: Andrey Shur
#       Latest update: 12/21/2020
#
################################################################


# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import copy
from warnings import warn

from .component import Component
from .components_basic import DNA, RNA
from .dna_part import DNA_part
from .dna_part_misc import IntegraseSite
from .species import ComplexSpecies, OrderedPolymer, OrderedPolymerSpecies
from .utils import all_comb, remove_bindloc
import logging

class Construct(Component,OrderedPolymer):
    def __init__(self,
                parts_list,
                name=None,
                circular=False,
                mechanisms=None,  # custom mechanisms
                parameters=None,  # customized parameters
                attributes=None,
                initial_concentration=None, 
                component_enumerators = None,
                make_dirless_hash = True,
                **keywords):
        """this represents a bunch of parts in a row.
        A parts list has [[part,direction],[part,direction],...]
        Each part must be an OrderedMonomer"""

        if(component_enumerators is None):
            component_enumerators = []
        self.component_enumerators = component_enumerators
        OrderedPolymer.__init__(self,parts_list, default_direction="forward")
        self.circular=circular
        if(name is None):
            name = self.make_name() #automatic naming
        self.name = name
        Component.__init__(self=self,name=name,length = len(parts_list),
                    mechanisms=mechanisms,parameters=parameters,
                    attributes=attributes,initial_concentration = initial_concentration,
                     **keywords)
        self.update_parameters()
        self.transcripts = []
        if(not hasattr(self,"material_type")):
            self.material_type=None #set this when you inherit this class
        self.out_components = None
        self.predicted_rnas = None
        self.predicted_proteins = None
        self.directionless_hash = None
        if(make_dirless_hash):
            self.update_permutation_hash()

    @property
    def parts_list(self):
        return self.polymer
    def make_name(self):
        output = ""
        outlst = []
        for part in self.parts_list:
            pname = part.name
            if(part.direction=="reverse"):
                pname+="_r"
            outlst += [pname]
        output = '_'.join(outlst)
        if(self.circular):
            output+="_o"
        return output

    def get_part(self,part = None, part_type=None, name = None, index = None):
        """
        Function to get parts from Construct.parts_list.

        One of the 3 keywords must not be None.

        part: an instance of a DNA_part. Searches Construct.parts_list for a DNA_part with the same type and name.
        part_type: a class of DNA_part. For example, Promoter. Searches Construct.parts_list for a DNA_part with the same type.
        name: str. Searches Construct.parts_list for a DNA_part with the same name
        index: int. returns Construct.parts_list[index]

        if nothing is found, returns None.
        """

        if [part, name, index,part_type].count(None) != 3:
            raise ValueError(f"get_component requires a single keyword. Recieved component={part}, name={name}, index={index}.")
        if not (isinstance(part, DNA_part) or part is None):
            raise ValueError(f"component must be of type DNA_part. Recieved {part}.")
        if not (type(part_type) == type or part_type is None):
            raise ValueError(f"part_type must be a type. Recieved {part_type}.")
        if not (isinstance(name, str) or name is None):
            raise ValueError(f"name must be of type str. Recieved {name}.")
        if not (isinstance(index, int) or index is None):
            raise ValueError(f"index must be of type int. Recieved {index}.")

        matches = []
        if index is not None:
            matches.append(self.parts_list[index])
        else:
            for comp in self.parts_list:
                if part is not None:
                    if type(part) == type(comp) and comp.name == part.name:
                        matches.append(comp)
                elif name is not None:
                    if comp.name == name:
                        matches.append(comp)
                elif part_type is not None:
                    if(isinstance(comp,part_type)):
                        matches.append(comp)
        if len(matches) == 0:
            return None
        elif len(matches) == 1:
            return matches[0]
        else:
            warn("get_part found multiple matching components. A list has been returned.")
            return matches 


    def reverse(self):
        """reverses everything, without actually changing the DNA.
        also updates the name and stuff, since this is now a different Construct"""
        OrderedPolymer.reverse(self)
        self.reset_stored_data()
        self.name = self.make_name()
        #self.update_base_species()
        return self
    def get_reversed(self):
        """returns a reversed version of this construct without changing this construct"""
        outcon = copy.deepcopy(self)
        outcon.reverse()
        return outcon
    def get_circularly_permuted(self,new_first_position):
        """returns a new construct which has the first position changed to new_first_position"""
        if(not self.circular):
            return ValueError("cannot circularly permute linear construct")
        else:
            return DNA_construct(self.parts_list[new_first_position:]+self.parts_list[:new_first_position], circular=True, 
                            component_enumerators = self.component_enumerators, attributes=self.attributes,mechanisms=self.mechanisms, mixture=self.mixture)
    def set_mixture(self, mixture):
        self.mixture = mixture
        for part in self.parts_list:
            part.set_mixture(mixture)
    def update_permutation_hash(self):
        """update the unique string generated to represent the content of this dna construct regardless of orientation and rotation"""
        self.directionless_hash,_,_ = Construct.omnihash(self)
    def update_base_species(self, base_name=None, attributes = None):
        if base_name is None:
            self.base_species = self.set_species(self.name, material_type = self.material_type, attributes = attributes)
        else:
            self.base_species = self.set_species(base_name, material_type = self.material_type, attributes = attributes)

    def update_parameters(self, overwrite_parameters = True):
        """update parameters of all parts in the construct"""
        Component.update_parameters(self = self,parameter_database=self.parameter_database)
        for part in self.parts_list:
            part.update_parameters(parameter_database = self.parameter_database,
                                    overwrite_parameters = overwrite_parameters)

    def add_mechanism(self, mechanism, mech_type = None, overwrite = False, optional_mechanism = False):
        Component.add_mechanism(self, mechanism, mech_type = mech_type, \
                                overwrite = overwrite, optional_mechanism = optional_mechanism)
        for part in self.parts_list:
            part.add_mechanism( mechanism, mech_type = mech_type, \
                                overwrite = overwrite, optional_mechanism = optional_mechanism)
    
    def __repr__(self):
        """this is just for display purposes"""
        return "Construct = "+ self.make_name()

    def __contains__(self,obj2):
        """checks if this construct contains a certain part, or a copy of a certain part"""
        if(isinstance(obj2,DNA_part)):
            #if we got a DNA part it could mean one of two things:
            #1 we want to know if a dna part is anywhere
            #2 we want to know if a specific DNA part is in here
            #this is complicated by the fact that we want to have the same DNA part be reusable
            #in many locations
            if(obj2.parent==self):
                #the object should already know if it's a part of me
                return True
            elif(obj2.parent==None):
                #this object has been orphaned. 
                #that means we are looking for matching objects in any position
                new_obj2 = copy.copy(obj2).unclone()
                uncloned_list = [copy.copy(a).unclone() for a in self.parts_list]
                return new_obj2 in uncloned_list
            else:
                return False
        elif(isinstance(obj2,str)):
            #if we get a string, that means we want to know if the name exists anywhere
            return obj2 in str(self)

    def get_species(self):
        """returns the species of this dna construct, using OrderedPolymerSpecies"""
        ocomplx = []
        for part in self.parts_list:
            partspec = copy.copy(part.dna_species)
            ocomplx += [partspec.set_dir(part.direction)]
        out_species = OrderedPolymerSpecies(ocomplx,circular = self.circular,material_type=self.material_type)
        
        return out_species
    
    def located_allcomb(self,spec_list):
        """recursively trace all paths through a list
        [[[part1,1],[part2,5]],[[part3,1]],[[part4,5],[part5,12]]]
        ====================>
        compacted_indexes = [1,5,12]
        prototype_list = [[part1,part3],[part2,part4],[part5]]
        comb_list = [[1],[5],[12],[1,5],[1,12],[5,12],[1,5,12]]
        ===========================
        then, take the lists from comb_list and create all possible lists
        out of prototype_list that includes those elements"""
        #first we have to construct the list we are tracing paths through
        
        spec_indexes = [a.position for a in spec_list] #extract all indexes
        #the following takes apart the lists because i don't yet know how to deal
        #with multiple binders at the same time
        compacted_indexes = sorted(list(set(spec_indexes)))

        prototype_list = [None]*len(compacted_indexes)
        for spec in spec_list:
            #now, spec is a list which contains all the elements which are defined for each variant.
            #
            #go through every element and put it in the right place
            proto_ind = compacted_indexes.index(spec.position) #where to put it?
            if(prototype_list[proto_ind] is None):
                #if nothing's been placed here, then create a list
                prototype_list[proto_ind] = [spec]
            else:
                #if something is already here, then add to the list
                prototype_list[proto_ind]+= [spec]
        # at this point we have a list that looks like this:
        # [[[part1,0],[part2,0]],[[part2,3]],[[part3,12],[part5,12]]
        # next step is to pick one of the first list (either [part1,0] or [part2,0])
        # one of the second list (only [part2,3] is our option), one of the third list, etc
        # for all possible choices made this way
        comb_list = all_comb(compacted_indexes)
        def recursive_path(in_list):
            """this function takes every possible "path" through a list of lists.
            For example:
            input:  [[A,B],[C],[D,E]]
            we have two options for the first position, one option for the second, and two options for the third.
            so, output:[[A,C,D],[B,C,D],[A,C,E],[B,C,E]]
            """
            if(len(in_list)==1):
                out_list = []
                for a in in_list[0]:
                    out_list+= [[a]]
                return out_list
            elif(len(in_list)==0):
                return []
            else:
                out_list = []
                for a in in_list[0]:
                    out_list += [[a] + z for z in recursive_path(in_list[1:])]
                return out_list
        outlist = []
        for combo in comb_list:
            combo_sublists = []
            for combo_index in combo:
                combo_sublists += [prototype_list[compacted_indexes.index(combo_index)]]
            outlist+= recursive_path(combo_sublists)
        outdict_list = []
        for combo in outlist:
            replacedict = {}
            for spec in combo:
                replacedict.update({spec.position:[spec,spec.direction]})
            outdict_list += [replacedict]
        return outdict_list

    def make_polymers(self,species_lists,backbone):
        """makes polymers from lists of species
        inputs:
        species_lists: list of species which are to be assembled into a polymer
        backbone: the base_species which all these polymers should have"""
        polymers = []
        for combo in species_lists:
            #to make combinatorially bound versions of a species, we take the unbound species (backbone)
            #and replace monomers of it with bound versions (combo)
            #combo is a dictionary of the form: {number:OrderedMonomer,...}
            #where the number is which element should be replaced and the OrderedMonomer is what will be replacing that element
            new_backbone = OrderedPolymerSpecies.from_polymer_species(backbone,combo)
            polymers += [new_backbone] #we make a new OrderedComplexSpecies
        return polymers

    def update_combinatorial_complexes(self,active_components):
        """given an input list of components, we produce all complexes
        yielded by those components, mixed and matched to make all possible combinatorial
        complexes, where each component is assumed to only care about binding to one spot.
        First, the components are asked what species they make, then these species are sifted to
        reveal only the ones which are versions of the same polymer, just with different locations bound.
        Then, combinatorial combinations are made.
        for example:
            construct: <A,B,C>
        two new species are possible: <[A:RNAP],B,C>; <A,[B:RNAP],C>
        combinatorial species is also possible (since A and B are assumed to act independantly)
                                        <[A:RNAP],[B:RNAP],C>
        complexes, where each component is assumed to only care about binding to one spot"""
        species = []
        for part in active_components:
            #first we make binary complexes
            sp_list =  part.update_species()
            species+=sp_list
        unique_complexes = []
        #species need to be uniqueified
        unique_species = list(set(species)) 
        for specie in unique_species:
            #in this list we extract all the variants of the complexes from possible_backbones
            #that exist in our species list.
            if(isinstance(specie,ComplexSpecies) and specie.position is not None):
                unique_complexes+=[specie]
        #unique_complexes now has a list of all the non-combinatorial complexes we can make
        combinatorial_complexes = unique_complexes
        allcomb = self.located_allcomb(combinatorial_complexes) #all possible combinations of binders are made here
        allcomb += [{}] #unbound dna should also be used
        #now, all possibilities have been enumerated.
        #we construct the OrderedPolymerSpecies
        out_polymers = self.make_polymers(allcomb,self.get_species())
        return out_polymers

    #Overwrite Component.enumerate_components 
    def enumerate_constructs(self,previously_enumerated=None):
        """Runs all our component enumerators to generate new constructs"""
        new_constructs = []
        for enumerator in self.component_enumerators:
            new_comp = enumerator.enumerate_components(component=self,previously_enumerated=previously_enumerated)
            new_constructs += new_comp
        return new_constructs

    def combinatorial_enumeration(self):
        """returns a list of new components that are copies of
        existing components, but with a different species placed inside. This different
        species represents different combinatorial states of the polymer.
        for example:
            construct: <A,B,C>
        two new species are possible: <[A:RNAP],B,C>; <A,[B:RNAP],C>
        combinatorial species is also possible (since A and B are assumed to act independantly)
                                        <[A:RNAP],[B:RNAP],C>
        Thus, this function returns A which binds to <A,B,C> (creating <[A:RNAP],B,C>) AND also
        A which binds to <A,[B:RNAP],C> (creating <[A:RNAP],[B:RNAP],C>). Likewise for B
        In total two A components are returned, and two B components are returned.
        """
        #Looks at combinatorial states of constructs to generate DNA_parts
        #my_polymer = self.get_species()
        self.update_parameters()

        #Go through parts
        active_components = []
        for part in self.parts_list:
            if(hasattr(part,"update_component")):
                dummy_species = self.get_species()[part.position]
                updated_components = part.update_component(dummy_species,practice_run=True)
                if(updated_components is not None):
                    active_components += [updated_components]
        #this next part creates "combinatorial" bound complexes, given singly bound complexes generated above.
        #for example, let's say we got <[A:RNAP],B,C> and <A,[B:RNAP],C> from A and B individually, then a combinatorial
        #construct would be <[A:RNAP],[B:RNAP],C>
        combinatorial_complexes = self.update_combinatorial_complexes(active_components)

        combinatorial_components = []
        for comb_specie in combinatorial_complexes:
            #after making all combinatorially bound parts, we must seed components with the right species
            #so that the right reactions are made. For example, A cannot react with <[A:RNAP],[B:RNAP],C> because
            #that's the species you get after A has already reacted. So here we are finding species with the
            #proper positions unbound, so they can be fed to the proper components
            if(isinstance(comb_specie,OrderedPolymerSpecies)):
                for part in active_components:
                    part_pos = part.position
                    if(not isinstance(comb_specie[part_pos],ComplexSpecies)):
                        #in this case the position of interest is not complexed. Thus, we need to update!
                        combinatorial_components += [part.update_component(comb_specie[part_pos])]
        return combinatorial_components

    def enumerate_components(self,previously_enumerated=None):
        """returns a list of new components and constructs.
        New components are generated if:
        - a component creates a species which results in binding to part of the construct
            Example: <A,B,C> -> <[A:RNAP],B,C>
            Then, A would be returned since a new species is created
        - more than one such component exist in the same construct, for example:
            construct: <A,B,C>
            two new species are possible: <[A:RNAP],B,C>; <A,[B:RNAP],C>
            combinatorial species is also possible (since A and B are assumed to act independantly)
                                         <[A:RNAP],[B:RNAP],C>
            Thus, this function returns A which binds to <A,B,C> (creating <[A:RNAP],B,C>) AND also
            A which binds to <A,[B:RNAP],C> (creating <[A:RNAP],[B:RNAP],C>). Likewise for B
            In total two A components are returned, and two B components are returned.
        New constructs are generated if:
            self.enumerate_construcs() says so.
            For example, in <A,B,C>, A is a promoter and makes an RNA_construct containing <B,C>
        """
        #Runs component enumerator to generate new constructs
        new_constructs = self.enumerate_constructs(previously_enumerated=previously_enumerated)

        #Looks at combinatorial states of constructs to generate DNA_parts
        combinatorial_components = self.combinatorial_enumeration()

        return combinatorial_components+new_constructs
    @classmethod
    def get_partstring(cls,part):
        """a string name of a part including its name and direction (and not position)"""
        orphan = part.get_orphan()
        orphan.direction = None
        orphan.position = None
        curname = str(orphan)
        curdir = part.direction
        return curname+curdir
    @classmethod
    def get_partlist_hash(cls,partlist):
        """creates a string containing the name and direction of all parts in a list of parts (but not their position)"""
        partlist_str = '_'.join([str(a[0])+str(b) for a,b in zip(partlist,range(len(partlist)))])
        return partlist_str
    @classmethod
    def create_hashless_reverse(cls,construct):
        """create a reverse construct but don't calculate its hash (because that would make an infinite loop)"""
        rev_con = [a.get_orphan() for a in construct]
        for rev_part,origpart in zip(rev_con,construct):
            rev_part.direction = {"forward":"reverse","reverse":"forward"}[origpart.direction]
        rev_con = rev_con[::-1]
        rev_con = Construct(rev_con,make_dirless_hash=False,circular = construct.circular)
        return rev_con
    @classmethod
    def rotation_free_hash(cls,construct): 
        """calculates a unique circular permutation that is the most alphabetically ordered. Every part is considered as a potential starting
        point, and the most alphabetically ordered order is then chosen as the best permutation
        
        returns:
        hash of the most alphabetical ordering, direction of the ordering (always 1), first position of the best rotation
        
        thus, to recreate the conformation of the construct used to make this hash you would have to 
        1) invert the construct (or not), then 
        2) use the indicated position as the first position"""
        def circular_next(part,construct):
            if(isinstance(part,int)):
                part_pos = part
            else:
                part_pos = part.position+1
            part_pos = part_pos%(len(construct))
            return construct[part_pos]
        best_partlist = [(None,None)]
        for part in construct:
            parthash = Construct.get_partstring(part)
            if(best_partlist[0][0] is None):
                best_partlist = [(parthash,part)]
            elif(parthash > best_partlist[0][0]):
                best_partlist = [(parthash,part)]
            elif(parthash == best_partlist[0][0]):
                test_partlist = [(parthash,part)]
                testpos = part.position

                while test_partlist[-1][0] == best_partlist[len(test_partlist)-1][0]:
                    if(testpos-part.position>= len(construct)):
                        #this means we tested every part and they would come in at the same alphabetical position
                        break
                    testpos += 1
                    next_testpart = circular_next(testpos,construct)
                    test_partlist += [(Construct.get_partstring(next_testpart),next_testpart)]

                    if(len(best_partlist)<len(test_partlist)):
                        #this means we haven't made the hash string for the next part down the line
                        next_part = circular_next(best_partlist[-1][1],construct)
                        next_parthash = Construct.get_partstring(next_part)
                        best_partlist += [(next_parthash,next_part)]
                
                if(test_partlist[-1][0] > best_partlist[len(test_partlist)-1][0]):
                    best_partlist = test_partlist
        while(len(best_partlist)<len(construct)):
            nextpart = circular_next(best_partlist[-1][1],construct)
            best_partlist += [(Construct.get_partstring(nextpart),nextpart)]
        return Construct.get_partlist_hash(best_partlist),1,best_partlist[0][1].position
    @classmethod
    def direction_rotation_free_hash(cls,construct):
        """computes the best circular permutation of a construct, forward and reverse. Then returns whichever one of those comes alphabetically first
        
        returns:
        hash of the most alphabetical ordering, direction of the ordering (1 or -1), first position of the best rotation
        
        thus, to recreate the conformation of the construct used to make this hash you would have to 
        1) invert the construct (or not), then 
        2) use the indicated position as the first position"""
        rev_con = Construct.create_hashless_reverse(construct)
        forward_hash,_,pos = Construct.rotation_free_hash(construct)
        reverse_hash,_,posrev = Construct.rotation_free_hash(rev_con)
        if(forward_hash>reverse_hash):
            return forward_hash,1,pos
        else:
            return reverse_hash,-1,posrev
    @classmethod
    def linear_direction_free_hash(cls,construct):
        """creates a string representing the construct forward or reverse, and returns whichever of those is alphabetically first
        
         returns:
        hash of the most alphabetical ordering, direction of the ordering (1 or -1), first position of the best rotation (always 0)
        
        thus, to recreate the conformation of the construct used to make this hash you would have to 
        1) invert the construct (or not), then 
        2) use the indicated position as the first position"""
        rev_con = Construct.create_hashless_reverse(construct)

        test_partlist = []
        test_reverse_partlist = []
        winner = None
        for part,revpart in zip(construct,rev_con):
            #test both forward and reverse at the same time
            if(winner is None):
                parthash = Construct.get_partstring(part)
                rev_parthash = Construct.get_partstring(revpart)
                if(parthash>rev_parthash):
                    winner = "forward"
                elif(parthash<rev_parthash):
                    winner = "reverse"
                test_partlist += [(parthash,part)]
                test_reverse_partlist += [(rev_parthash,revpart)]
            elif(winner=="forward"):
                #as soon as one of the two directions comes out on top just go with that one
                test_partlist += [(Construct.get_partstring(part),part)]
            elif(winner=="reverse"):
                #as soon as one of the two directions comes out on top just go with that one
                test_reverse_partlist += [(Construct.get_partstring(revpart),revpart)]
        if(winner=="forward"):
            rethash = Construct.get_partlist_hash(test_partlist)
            return rethash,1,0
        elif(winner == "reverse"):
            rethash = Construct.get_partlist_hash(test_reverse_partlist)
            return rethash,-1,0
    @classmethod
    def omnihash(cls,construct):
        """a construct can exist forwards or backwards, and circularly permuted (but only if it's a circular construct). 
        This function creates the "best" circular permutation and ordering of a construct. But the circular permutation 
        is only calculated if the construct is circular. Best is calculated based on which orientation/ permutation has 
        the most part names in alphabetical order.
        
        returns:
        hash of the most alphabetical ordering (string), direction of the ordering (1 or -1), first position of the best rotation
        
        thus, to recreate the conformation of the construct used to make this hash you would have to 
        1) invert the construct (or not), then 
        2) use the indicated position as the first position"""
        if(construct.circular):
            rhash,flip,posit = Construct.direction_rotation_free_hash(construct)
            return rhash+"circular",flip,posit
        else:
            rhash,flip,posit = Construct.linear_direction_free_hash(construct)
            return rhash+"linear",flip,posit
    def __hash__(self):
        return OrderedPolymer.__hash__(self)
    def __eq__(self,construct2):
        """equality means comparing the parts list in a way that is not too deep"""
        #TODO: make this be a python object comparison
        if((self.__repr__()==construct2.__repr__()) and (self.name == construct2.name)):
            return True
        else:
            return False
    
    def update_species(self):
        species = [self.get_species()]
        return species
    def reset_stored_data(self):
        self.out_components = None
        self.predicted_rnas = None
        self.predicted_proteins = None
    def changed(self):
        self.reset_stored_data()
        self.name = self.make_name()
    def update_reactions(self,norna=False):
        return []
        


class DNA_construct(Construct,DNA):
    def __init__(self,
                parts_list,
                name=None,
                circular=False,
                mechanisms=None,  # custom mechanisms
                parameters=None,  # customized parameters
                attributes=None,
                initial_concentration=None,
                copy_parts=True,
                component_enumerators = None,
                **keywords):

        self.material_type = "dna"
        if component_enumerators is None:
            from .construct_explorer import TxExplorer
            component_enumerators = [TxExplorer()]
        
        Construct.__init__(self=self, parts_list =parts_list, name = name, \
                            circular=circular, mechanisms=mechanisms, \
                            parameters=parameters, attributes=attributes, \
                            initial_concentration=initial_concentration, \
                            component_enumerators = component_enumerators, **keywords)
        DNA.__init__(self=self,name=self.name)
    def __repr__(self):
        return "DNA_construct = "+ self.make_name()
    
        
class RNA_construct(Construct,RNA):
    def __init__(self,parts_list,name=None,promoter=None,\
                component_enumerators = None,\
                **keywords):
        """an RNA_construct is a lot like a DNA_construct except it can only translate, and
        can only be linear"""
        
        self.material_type = "rna"
        self.promoter = promoter
        if component_enumerators is None:
            from .construct_explorer import TlExplorer
            component_enumerators = [TlExplorer()]

        Construct.__init__(self=self,parts_list=parts_list,circular=False,name=name,\
                                component_enumerators = component_enumerators,**keywords)
        RNA.__init__(self=self,name=self.name)
    def __repr__(self):
        """the name of an RNA should be different from DNA, right?"""
        return "RNA_construct = "+self.name
        



# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

from typing import List
from warnings import warn


class Mechanism(object):
    """Mechanism class for core mechanisms.

    Core mechanisms within a mixture (transcription, translation, etc)

    The Mechanism class is used to implement different core
    mechanisms in TX-TL.  All specific core mechanisms should be
    derived from this class.

    """
    def __init__(self, name: str, mechanism_type=""):
        """Initializes a Mechanism instance.

        :param name: name of the Mechanism
        :param mechanism_type: mechanism_type in string
        """
        self.name = name
        self.mechanism_type = mechanism_type
        if mechanism_type == "" or mechanism_type is None:
            warn(f"Mechanism {name} instantiated without a type. This could "
                 "prevent the mechanism from being inherited properly.")

    def update_species(self, component=None, part_id=None) -> List:
        """the child class should implement this method.

        :return: empty list
        """
        warn(f"Default Update Species Called for Mechanism = {self.name}.")
        return []

    def update_reactions(self, component=None, part_id=None) -> List:
        """ the child class should implement this method.

        :return: empty list
        """
        warn(f"Default Update Reactions Called for Mechanism = {self.name}.")
        return []

    def __repr__(self):
        return self.name


class EmptyMechanism(Mechanism):
    """For use when one needs a Mechanism to do nothing, such as translation in Expression Mixtures."""
    def __init__(self, name, mechanism_type):
        """Initializes an EmptyMechanism instance.

        :param name: name of the Mechanism
        :param mechanism_type: mechanism_type in string
        """
        Mechanism.__init__(self, name=name, mechanism_type=mechanism_type)

    def update_species(self, component=None, part_id=None, **keywords):
        return []

    def update_reactions(self, component=None, part_id=None, **keywords):
        return []




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import copy
from warnings import warn

from .dna_part import DNA_part
from .mechanisms_binding import One_Step_Cooperative_Binding
from .species import Species, ComplexSpecies, Complex
from .mechanisms_integrase import BasicIntegration
from .component import Component
from .components_basic import DNA

integrase_sites = ["attB","attP","attL","attR","FLP","CRE"]
class DNABindingSite(DNA_part):
    def __init__(self,name,binders,no_stop_codons=None,assembly = None,**keywords):
        """an integrase attachment site binds to integrase"""

        if(isinstance(binders,list)):
            self.binders = [self.set_species(a) for a in binders]
        else:
            self.binders = [binders]
        self.mechanisms = {"binding":One_Step_Cooperative_Binding()}
        DNA_part.__init__(self,name,no_stop_codons=no_stop_codons,mechanisms = self.mechanisms,assembly = assembly,**keywords)
        self.name = name
        self.dna_to_bind = None
        #self.assembly = None
        
    def __repr__(self):
        myname = self.name
        return myname
    def update_species(self):
        spec = []
        spec += self.binders
        if(self.dna_to_bind is not None):
            mech_b = self.mechanisms["binding"]
            for binder in self.binders:
                spec += mech_b.update_species(binder,self.dna_to_bind,component=self,part_id = binder.name)
                #TODO: different proteins probably have different affinity to bind this sequence
        return spec
    def update_reactions(self):
        rxns = []
        if(self.dna_to_bind is not None):
            mech_b = self.mechanisms["binding"]
            for binder in self.binders:
                rxns += mech_b.update_reactions(binder,self.dna_to_bind,component=self,part_id = binder.name)
        return rxns
    def update_component(self,internal_species=None,**keywords):
        """returns a copy of this component, except with the proper fields updated"""
        if(isinstance(self.parent,DNA)):
            out_component = copy.copy(self)
            out_component.dna_to_bind = internal_species
            return out_component
        else:
            return None
class IntegraseSite(DNABindingSite):
    def __init__(self,name, site_type = "attB",integrase = "int1", dinucleotide = 1,no_stop_codons=None,integrase_binding=True,**keywords):
        self.update_integrase(integrase)
        #self.integrase = integrase
        self.dinucleotide = dinucleotide
        self.site_type = site_type
        self.other_dna = None
        self.linked_sites = {}
        self.complexed_version = None
        self.integrase_binding=integrase_binding
        DNABindingSite.__init__(self,name,self.integrase,no_stop_codons=no_stop_codons,**keywords)
        self.add_mechanism(BasicIntegration(self.integrase.name))
    def __repr__(self):
        myname = self.name
        if(self.site_type in integrase_sites):
            myname += "_" + self.integrase.name
            if(self.dinucleotide != 1):
                myname += "_"+str(self.dinucleotide) 
        else:
            warn("warning! site {} has site_type {} which is not recognized".format(self.name,self.site_type))
        if(self.position is not None):
            myname+="_"+str(self.position)
        if(self.direction is not None):
            myname+="_"+str(self.direction[0])
        return myname
    def update_integrase(self,int_name):
        self.integrase = Component.set_species(int_name,material_type='protein')
    def __hash__(self):
        sumhash = DNABindingSite.__hash__(self) + self.dna_to_bind.__hash__()
        return sumhash
    def get_complexed_species(self,dna):
        recomp = Complex([dna,self.integrase,self.integrase])
        return recomp
    def update_component(self,internal_species=None,**keywords):
        """returns a copy of this component, except with the proper fields updated"""
        newcomp = DNABindingSite.update_component(self,internal_species=internal_species)
        #above is updating the component to take into account integrase binding (the default feature of DNABindingSite)
        if(newcomp is None):
            #if nothing binds, then nothing else happens. Since, integrase must be bound in order for integrase sites to do anything
            return None
        elif("practice_run" in keywords and keywords["practice_run"]):
            #combinatorial enumeration calls update_component twice.
            #an integrase site must inform all the sites it is linked to that it has been
            #updated. In certain cases the status of whether a site has been updated is informative
            #intramolecular sites only output reactions if they haven't been updated
            #intermolecular sites only output reactions if they have been updated
            #a site can be both types, it depends on the contents of self.linked_sites[site]
            
            #however, during the practice run we are trying to preserve the "initial" configuration
            #that the sites get after they are first created.
            #schematic:
            #site1 <===> site2
            #  ||        /\
            #update      /
            #  ||       /
            # \../     /
            #copy(site1) 
            # 
            #this site is linked to site2 but site2 is not linked to the copy
            #this is what we are trying to fix here
            for othersite_tpl in self.linked_sites:
                othersite = othersite_tpl[0]
                intermolecular = othersite_tpl[1]
                #what we are doing here is swapping out the link to this site with
                #a link to the returned component (the copied site)
                self_tuple = (self,intermolecular)
                mystuff = copy.copy(othersite.linked_sites[self_tuple])
                del othersite.linked_sites[self_tuple]
                othersite.linked_sites[(newcomp,intermolecular)] = mystuff
            return newcomp
        else:
            for othersite_tpl in self.linked_sites:
                othersite = othersite_tpl[0]
                intermolecular = othersite_tpl[1]
                populate = True #by default, add the appropriate data members to the other site
                if(not intermolecular):
                    #the reaction with the site in question is intramolecular
                    #that means we should only populate the other site if our internal_species
                    #has the proper location bound by integrase
                    #if the reaction is intramolecular, that means that this component knows everything in order to decide
                    #create the reaction.
                    #the linked site is populated because then the linked site knows not to create the reaction.
                    #after all, there is one reaction per two sites
                    
                    if(self.integrase_binding):
                        otherisbound = othersite.integrase in internal_species.parent[othersite.position]
                    else:
                        otherisbound = not isinstance(internal_species.parent[othersite.position],ComplexSpecies)
                    if(not otherisbound):
                        populate = False
                if(populate):
                    #if the reaction is intermolecular then the linked site is populated.
                    #this means only a fully populated site would have all the information to create the reaction
                    #once again this results in one out of two sites that actually outputs a "reaction" object, as required
                    mystuff = copy.copy(othersite.linked_sites[(self,intermolecular)])
                    del othersite.linked_sites[(self,intermolecular)]
                    othersite.linked_sites[(newcomp,intermolecular)]=mystuff
                    assert((othersite,intermolecular) in newcomp.linked_sites)
            return newcomp
    def update_species(self):
        if(self.integrase_binding):
            return DNABindingSite.update_species(self)
        else:
            return self.binders
    def update_reactions(self):
        if(self.integrase_binding):
            reactions = DNABindingSite.update_reactions(self)
        else:
            reactions = []
        if(self.linked_sites == {}):
            return reactions
        complex_parent = self.get_complexed_species(self.dna_to_bind).parent
        if(not self.integrase_binding):
            complex_parent = self.dna_to_bind.parent
        int_mech = self.get_mechanism('integration')

        #this next part generates integrase reactions
        for site_tpl in self.linked_sites:
            site = site_tpl[0]
            intermolecular = site_tpl[1]
            if(site.dna_to_bind is None or site.dna_to_bind.parent is None):
                #skip sites which do not know who they bind to
                continue
            integrated_dnas = []
            #each integrase reaction will have an entry here
            
            #now we must know if the site pair got processed. If it did,
            #then generate the reaction. If it didn't then update the pair

            #however, if we are dealing with an intramolecular reaction, then
            #only return the reaction if the pair HASN'T been processed
            populate = True
            if(not intermolecular):
                #this is an intramolecular reaction so we don't care about the other site's DNAs
                integrated_dnas = []
                site_is_bound = site.integrase in complex_parent[site.position] #site is bound by integrase
                site_isnt_bound = not isinstance(complex_parent[site.position],ComplexSpecies) #site isnt bound by anything
                if(( (self.integrase_binding and site_is_bound) or (not self.integrase_binding and site_isnt_bound)) 
                                                                    and complex_parent in self.linked_sites[site_tpl][1]):
                    #make sure that both this site and the other site are bound, or not depending on the value of "integrase_binding"
                    for integrase_function in self.linked_sites[site_tpl][0]:
                        integr = integrase_function.create_polymer([complex_parent])
                        integrated_dnas += [integr]
                    reactions += int_mech.update_reactions([complex_parent],integrated_dnas,component=self,part_id = self.integrase.name)
                    populate = False
            else:
                #this is for intermolecular reactions. now "other_dna" is possible
                #go through all possible "other" dnas then calculate for each RNA
                for other_dna in self.linked_sites[site_tpl][1]:
                    #other_dna are placed there by other sites that already got evaluated
                    integrated_dnas = []
                    for integrase_function in self.linked_sites[site_tpl][0]:
                        #for every result that we could get, generate it
                        #the next line generates the OrderedPolymerSpecies which results from recombination
                        integrated_dnas += [integrase_function.create_polymer([complex_parent,other_dna])]
                    
                    reactions += int_mech.update_reactions([complex_parent,other_dna],integrated_dnas,component=self,part_id = self.integrase.name)
            #next part updates the linked site
            if(populate and (complex_parent is not None) and (self,intermolecular) in site.linked_sites and\
                 (complex_parent not in site.linked_sites[(self,intermolecular)][1])):
                site.linked_sites[(self,intermolecular)][1] += [complex_parent]
        return reactions

class UserDefined(DNA_part):
    def __init__(self,name,dpl_type=None, **keywords):
        """a user defined part is a part that doesn't do anything, 
        just exists as a label basically"""
        DNA_part.__init__(self,name, **keywords)
        self.dpl_type = dpl_type
        self.name = name
    def update_species(self):
        return []
    def update_reactions(self):
        return []

class Origin(DNA_part):
    def __init__(self,name, **keywords):
        """an origin does nothing except look right when plotted"""
        DNA_part.__init__(self,name, **keywords)
        self.name = name
    def update_species(self):
        return []
    def update_reactions(self):
        return []
class Operator(DNA_part):
    def __init__(self,name,binder=None, **keywords):
        """an operator does nothing except look right when plotted"""
        DNA_part.__init__(self,name, **keywords)
        self.binder = []
        if(binder is not None):
            for bind in binder:
                self.binder += [Component.set_species(bind)]
        self.name = name
    def update_species(self):
        return []
    def update_reactions(self):
        return []


from .mechanism import Mechanism
from .reaction import Reaction
from .species import Complex
from .propensities import HillPositive


# [precursors] --> [products] using Massaction (None OK)
class OneStepPathway(Mechanism):
    def __init__(self, name = "one_step_pathway", mechanism_type = "metabolic_pathway"):
        Mechanism.__init__(self, name = name, mechanism_type = mechanism_type)


    def update_species(self, precursor, product, **keywords):
        species = []
        if precursor is not None:
            species += precursor
        if product is not None:
            species += product
        return species

    def update_reactions(self, precursor, product, component = None, part_id = None, k = None, **keywords):
        if precursor is None:
            inputs = []
        else:
            inputs = precursor

        if product is None:
            outputs = []
        else:
            outputs = product

        if component is None and k is None:
            raise ValueError("Must pass in a component or a rate k.")
        elif k is None:
            k = component.get_parameter("k", part_id = part_id, mechanism = self)

        r = Reaction.from_massaction(inputs, outputs, k_forward = k)
        return [r]



